#include <sys/types.h>
#include <sys/module.h>
#include <sys/systm.h>  /* uprintf */
#include <sys/errno.h>
#include <sys/param.h>  /* defines used in kernel.h */
#include <sys/kernel.h> /* types used in module initialization */
#include <sys/conf.h>   /* cdevsw struct */
#include <sys/uio.h>    /* uio struct */
#include <sys/malloc.h>
#include <sys/condvar.h>
#include <sys/queue.h>
#include <sys/lock.h>
#include <sys/sx.h>
#include <sys/mutex.h>
#include <sys/proc.h>

#include <sys/mount.h> /* used by fdisp* fncs */
#include <sys/vnode.h> /* and v(fs|n)ops, too */

#include <sys/namei.h> /* v(fs|n)ops */
#include <sys/stat.h>
#include <sys/unistd.h> /* pathconf */
#include <sys/filedesc.h> /* open */
#include <sys/file.h> /* open */
#include <sys/fcntl.h> /* open */
#include <sys/dirent.h> /* readdir */

#include <sys/buf.h>
#include <sys/bio.h>
#include <vm/vm.h>
#include <vm/vm_extern.h>  
#include <vm/vnode_pager.h>

#include <vm/vm_object.h>

#include "fuse.h"

#if _DEBUG2G
#include <sys/kdb.h>
#endif

#if FMASTER
#define __static
#else
#define __static static
#endif

#define NOT_YET_USED 0

MALLOC_DEFINE(M_FUSEMSG, "fuse messaging",
              "buffer for fuse messaging related things");

uint32_t fuse_useco = 0;
#if USE_FUSE_LOCK
struct mtx fuse_mtx;
#endif

__static struct clonedevs *fuseclones;

static d_open_t      fusedev_open;
static d_close_t     fusedev_close;
static d_read_t      fusedev_read;
static d_write_t     fusedev_write;

static struct cdevsw fuse_cdevsw = {
	.d_open = fusedev_open,
	.d_close = fusedev_close,
	.d_name = "fuse",
	.d_read = fusedev_read,
	.d_write = fusedev_write,
	.d_version = D_VERSION,
#if ! DO_GIANT_MANUALLY
	.d_flags = D_NEEDGIANT,
#endif
};


#if USE_OLD_CLONEHANDLER_API
static void			 fusedev_clone(void *arg, char *name, int namelen, struct cdev **dev);
#else
static void			 fusedev_clone(void *arg, struct ucred *cred, char *name, int namelen, struct cdev **dev);
#endif
static void			 fuse_bringdown(eventhandler_tag eh_tag);
static int			 fuse_loader(struct module *m, int what, void *arg);

static __inline void		 fuse_iov_init(struct fuse_iov *fiov, size_t size);
static void			 fuse_iov_teardown(struct fuse_iov *fiov);
__static void			 fuse_iov_adjust(struct fuse_iov *fiov, size_t size);
static __inline void		 fuse_iov_audit(struct fuse_iov *fiov);
static void			 fuse_callbn_init(struct fuse_callback_node *caliban, struct fuse_ticket *tick);
static void			 fuse_callbn_teardown(struct fuse_callback_node *caliban);
static __inline void		 fuse_callbn_refresh(struct fuse_callback_node *caliban);
#if NOT_YET_USED
static __inline struct fuse_iov	*fuse_callbn_params(struct fuse_callback_node *caliban);
#endif
__static __inline struct fuse_iov	*fuse_callbn_resp(struct fuse_callback_node *caliban);
static __inline int		 fuse_callbn_answered(struct fuse_callback_node *caliban);
static __inline void		 fuse_callbn_set_answered(struct fuse_callback_node *caliban);
#if FMASTER
__static int			 fuse_callbn_wait_answer(struct fuse_callback_node *caliban);
#endif
__static __inline int		 fuse_callbn_pull_uio(struct fuse_callback_node *caliban, struct uio *uio);
__static __inline int		 fuse_callbn_get_opcode(struct fuse_callback_node *caliban);

static __inline void		 fuse_msgn_init(struct fuse_msg_node *msgn, struct fuse_ticket *tick, size_t size);
static __inline void		 fuse_msgn_teardown(struct fuse_msg_node *msgn);
static __inline void		 fuse_msgn_refresh(struct fuse_msg_node *msgn);

static struct fuse_ticket	*ticket_alloc(struct fuse_data *data);
static __inline void		 ticket_refresh(struct fuse_ticket *tick);
static void			 ticket_destroy(struct fuse_ticket *tick);
__static void			 ticket_drop(struct fuse_ticket *tick);
__static struct fuse_ticket 	*ticket_fetch(struct fuse_data *data);
__static __inline void		 ticket_invalidate(struct fuse_ticket *tick);
static __inline void		 ticket_drop_invalid(struct fuse_ticket *tick);
static __inline void		 fdata_kick_set(struct fuse_data *data);
static __inline int		 fdata_kick_get(struct fuse_data *data);

__static void			 fuse_insert_callback(struct fuse_ticket *tick, fuse_handler_t *handler);
__static void		 	 fuse_insert_message(struct fuse_ticket *tick);
#if FMASTER
__static fuse_handler_t fuse_standard_handler;
#endif

static struct fuse_data				*fdata_alloc(struct ucred *cred);
__static void					 fdata_destroy(struct fuse_data *data);
/**/
static __inline void				 fuse_msgn_push(struct fuse_msg_node *msgn);
static __inline struct fuse_msg_node		*fdata_pop_msg(struct fuse_data *data);
/**/
static __inline void				 fuse_callbn_push(struct fuse_callback_node *callbn);
static __inline void				 fuse_callbn_remove(struct fuse_callback_node *callbn);
#if NOT_YET_USED
static __inline struct fuse_callback_node	*fdata_pop_callb(struct fuse_data *data);
#endif
/**/
static __inline void				 ticket_push_ftick(struct fuse_ticket *tick);
static __inline struct fuse_ticket		*fdata_pop_ftick(struct fuse_data *data);
/**/
static __inline void				 ticket_push_btick(struct fuse_ticket *tick);
static __inline void				 ticket_remove_btick(struct fuse_ticket *tick);
static __inline struct fuse_ticket		*fdata_pop_btick(struct fuse_data *data);

static __inline int				 fuse_ohead_audit(struct fuse_out_header *ohead, struct uio *uio);
__static int					 fuse_body_audit(struct fuse_callback_node *caliban, size_t blen);
static __inline void				 fuse_setup_ihead(struct fuse_in_header *ihead,
                    				                  struct fuse_ticket *tick, uint64_t nid,
                    				                  enum fuse_opcode op, size_t blen,
                    				                  struct thread* td, struct ucred *cred);
__static __inline struct sx 			*fusedev_get_lock(struct cdev *fdev);
__static __inline struct fuse_data 		*fusedev_get_data(struct cdev *fdev);

static int		 fdisp_get_vopdata(struct fuse_dispatcher *fdip, struct mount *mp);
static void		 fdisp_prepare_msg(struct fuse_dispatcher *fdip,
           		                   enum fuse_opcode op, uint64_t nid,
           		                   struct thread *td, struct ucred *cred);
static __inline int	 fdisp_prepare_all(struct fuse_dispatcher *fdip,
                   	                   enum fuse_opcode op, struct vnode *vp,
                   	                   struct thread *td, struct ucred *cred);
__static int		 fdisp_wait_answ(struct fuse_dispatcher *fdip);
static __inline int	 fdisp_simple_putget(struct fuse_dispatcher *fdip,
                   	                     enum fuse_opcode op, struct vnode *vp,
                   	                     struct thread *td, struct ucred *cred);
static __inline int	 fuse_send_init(struct fuse_data *data, struct thread *td);
static __inline int	 fuse_send_forget(struct mount *mp, struct thread *td,
                   	                  struct ucred *cred, uint64_t nodeid,
                   	                  uint64_t nlookup,
                   	                 
                   	                  struct fuse_dispatcher *fdip);


/***************************************
 *
 * >>> Interface to objects (aka. "methods")
 *
 ***************************************/

/* fuse_iov methods ==> */

static void
fuse_iov_init(struct fuse_iov *fiov, size_t size)
{
	fiov->len = 0;
	fiov->base = malloc(FU_AT_LEAST(size), M_FUSEMSG, M_WAITOK | M_ZERO);
	fiov->allocated_size = FU_AT_LEAST(size);
}

static void
fuse_iov_teardown(struct fuse_iov *fiov)
{
	FREE(fiov->base, M_FUSEMSG);
}

__static __inline void
fuse_iov_adjust(struct fuse_iov *fiov, size_t size)
{
	/* XXX limits should be settable via sysctl */
	if (fiov->allocated_size < size ||
	    (FUSE_MAX_PERMANENT_IOBUF &&
	     fiov->allocated_size - size > FUSE_MAX_PERMANENT_IOBUF)) {
		fiov->base = realloc(fiov->base, FU_AT_LEAST(size), M_FUSEMSG,
		                     M_WAITOK | M_ZERO);
		fiov->allocated_size = FU_AT_LEAST(size);
	}

	fiov->len = size;
}

static __inline void
fuse_iov_audit(struct fuse_iov *fiov)
{
	bzero(fiov->base, fiov->len);	
	fuse_iov_adjust(fiov, 0);
}

/* <== fuse_iov methods */

/* fuse_msg_node methods ==> */

static __inline void
fuse_msgn_init(struct fuse_msg_node *msgn, struct fuse_ticket *tick, size_t size)
{
	msgn->ticket = tick;
	fuse_iov_init(&msgn->msg, size);
	msgn->msg_bufdata = NULL;
	msgn->msg_bufsize = 0;
}

static __inline void
fuse_msgn_teardown(struct fuse_msg_node *msgn)
{
	fuse_iov_teardown(&msgn->msg);
	msgn->msg_bufdata = NULL;
	msgn->msg_bufsize = 0;
}

static __inline void
fuse_msgn_refresh(struct fuse_msg_node *msgn)
{
	fuse_iov_audit(&msgn->msg);
	msgn->msg_bufdata = NULL;
	msgn->msg_bufsize = 0;
}

/* <== fuse_msg_node methods */

/* fuse_callback_node methods ==> */

static __inline void
fuse_callbn_init(struct fuse_callback_node *caliban, struct fuse_ticket *tick)
{
	caliban->ticket = tick;
	/*   -- These guys are allocated with M_ZERO
	caliban->ohead = NULL;
	caliban->errno = 0;
	*/
	mtx_init(&caliban->answer_mtx, "mutex for fuse answer delivery", NULL, MTX_DEF);
	fuse_iov_init(&caliban->params, 0);
	fuse_iov_init(&caliban->resp, 0);
	caliban->resp_bufdata = NULL;
	caliban->resp_bufsize = 0;
}

static void
fuse_callbn_teardown(struct fuse_callback_node *caliban)
{

	/* 
	 * It is important to use NFREE for freeing
	 * pointer elements of a fuse_callback_node
	 * when they are in use (and being dropped),
	 * so that when the end of its lifecycle comes,
	 * we can know whether those elements should be
	 * freed
	 */
	
	fuse_iov_teardown(&caliban->params);
	mtx_destroy(&caliban->answer_mtx);
	NFREE(caliban->ohead, M_TEMP);
	fuse_iov_teardown(&caliban->resp);
	caliban->resp_bufdata = NULL;
	caliban->resp_bufsize = 0;
}

static __inline void
fuse_callbn_refresh(struct fuse_callback_node *caliban)
{
	fuse_iov_audit(&caliban->params);
	NFREE(caliban->ohead, M_TEMP);
	fuse_iov_audit(&caliban->resp);
	caliban->errno = 0;
	caliban->resp_bufdata = NULL;
	caliban->resp_bufsize = 0;
}

#if NOT_YET_USED
static __inline struct fuse_iov *
fuse_callbn_params(struct fuse_callback_node *caliban)
{
	return (&caliban->params);
}
#endif

__static __inline struct fuse_iov *
fuse_callbn_resp(struct fuse_callback_node *caliban)
{
	return (&caliban->resp);
}

static __inline int
fuse_callbn_answered(struct fuse_callback_node *caliban)
{
	return (caliban->ticket->miscflag & FU_ANSW);
}

static __inline void
fuse_callbn_set_answered(struct fuse_callback_node *caliban)
{
	caliban->ticket->miscflag |= FU_ANSW;
}

__static int
fuse_callbn_wait_answer(struct fuse_callback_node *caliban)
{
	int err = 0;

	mtx_lock(&caliban->answer_mtx);
	if (fuse_callbn_answered(caliban))
		goto out;

	err = msleep(caliban->ticket, &caliban->answer_mtx, PCATCH, "fu ans", 0);

out:	
	mtx_unlock(&caliban->answer_mtx);

	if (! (err || fuse_callbn_answered(caliban))) {
		/*
		 * Some deadlocky scenarios can get us here, like SIGKILLing
		 * the fusexmp daemon after a fuse dir has been null mounted
		 * over its original copy in the "normal" fs
		 *
		 * (I guess there is no need of kicking the daemon at this
		 * point...)
		 */
	        DEBUG2G("fuse requester was woken up but still no answer");
		err = ENXIO;
	}

	return (err);
}

__static __inline int
fuse_callbn_pull_uio(struct fuse_callback_node *caliban, struct uio *uio)
{
	int err = 0;
	size_t len;

	len = uio->uio_resid;

	if (len) {
		if (caliban->resp_bufdata) {
			caliban->resp_bufsize = len;
			err = uiomove(caliban->resp_bufdata, len, uio);
		} else {
			fuse_iov_adjust(fuse_callbn_resp(caliban), len);
			err = uiomove(fuse_callbn_resp(caliban)->base, len, uio);
		}
	}
	DEBUG("uio pulled, error val: %d\n",err);
	return (err);
}

__static __inline int
fuse_callbn_get_opcode(struct fuse_callback_node *caliban)
{
	return (((struct fuse_in_header *)(caliban->ticket->msgn.msg.base))->opcode);
}

/* <== fuse_callback_node methods */

/* misc methods for managing queues and alike ==> */

static __inline void
fuse_msgn_push(struct fuse_msg_node *msgn)
{
	STAILQ_INSERT_TAIL(&msgn->ticket->data->fmsg_head, msgn, fmsg_link);
}

static __inline struct fuse_msg_node *
fdata_pop_msg(struct fuse_data *data)
{
 	struct fuse_msg_node *msgn;

	if ((msgn = STAILQ_FIRST(&data->fmsg_head)))
		STAILQ_REMOVE_HEAD(&data->fmsg_head, fmsg_link);
	return (msgn);
}
/**/
static __inline void
fuse_callbn_push(struct fuse_callback_node *callbn)
{
	TAILQ_INSERT_TAIL(&callbn->ticket->data->fcallb_head, callbn, fcallb_link);
}

static __inline void
fuse_callbn_remove(struct fuse_callback_node *callbn)
{
	TAILQ_REMOVE(&callbn->ticket->data->fcallb_head, callbn, fcallb_link);
}

#if NOT_YET_USED
static __inline struct fuse_callback_node *
fdata_pop_callb(struct fuse_data *data)
{
	struct fuse_callback_node *callbn;

	if ((callbn = TAILQ_FIRST(&data->fcallb_head)))
		fuse_callbn_remove(callbn);
	return (callbn);
}
#endif
/**/
static __inline void
ticket_push_ftick(struct fuse_ticket *tick)
{
	STAILQ_INSERT_TAIL(&tick->data->freetickets_head, tick, freetickets_link);
	tick->data->freeticket_counter++;
}

static __inline struct fuse_ticket *
fdata_pop_ftick(struct fuse_data *data)
{
	struct fuse_ticket *tick;

	if ((tick = STAILQ_FIRST(&data->freetickets_head))) {
		STAILQ_REMOVE_HEAD(&data->freetickets_head, freetickets_link);
		data->freeticket_counter--;
	}

	return tick;
}
/**/
static __inline void
ticket_push_btick(struct fuse_ticket *tick)
{
	TAILQ_INSERT_TAIL(&tick->data->busytickets_head, tick, busytickets_link);
}

static __inline void
ticket_remove_btick(struct fuse_ticket *tick)
{
	TAILQ_REMOVE(&tick->data->busytickets_head, tick, busytickets_link);
}

static __inline struct fuse_ticket *
fdata_pop_btick(struct fuse_data *data)
{
	struct fuse_ticket *tick;

	if ((tick = TAILQ_FIRST(&data->busytickets_head))) {
		ticket_remove_btick(tick);
	}

	return (tick);
}

/* <== misc methods for managing queues and alike */

/* fuse_data methods ==> */

static struct fuse_data *
fdata_alloc(struct ucred *cred)
{
	struct fuse_data *data;

	MALLOC(data, struct fuse_data *, sizeof(struct fuse_data), M_FUSEMSG,
	       M_WAITOK | M_ZERO);

	/* Setting up fields of mine */
	mtx_init(&data->msg_mtx, "mutex for fuse message list", NULL, MTX_DEF);
	STAILQ_INIT(&data->fmsg_head);
	mtx_init(&data->ticket_mtx, "mutex for the fuse ticketer", NULL,
	         MTX_DEF);
	STAILQ_INIT(&data->freetickets_head);
	TAILQ_INIT(&data->busytickets_head);
	mtx_init(&data->callback_mtx, "mutex for the fuse callback list", NULL,
	         MTX_DEF);
	TAILQ_INIT(&data->fcallb_head);
	data->ticketer = 0;
	data->freeticket_counter = 0;
	data->daemoncred = crhold(cred);

	sx_init(&data->slock, "lock protecting fuse_data");

	return (data);
}

__static void
fdata_destroy(struct fuse_data *data)
{
	struct fuse_ticket *tick;

	/* Driving off stage all that stuff thrown at device... */
	mtx_destroy(&data->msg_mtx);
	mtx_destroy(&data->callback_mtx);
	mtx_destroy(&data->ticket_mtx);

	while ((tick = fdata_pop_ftick(data)))
		ticket_destroy(tick);
		
	while ((tick = fdata_pop_btick(data)))
		ticket_destroy(tick);

	crfree(data->daemoncred);

	sx_destroy(&data->slock);

	FREE(data,M_FUSEMSG);
}

static __inline void
fdata_kick_set(struct fuse_data *data)
{
	DEBUG2G("banning daemon\n");
	mtx_lock(&data->msg_mtx);
	data->dataflag |= FDAT_KICK;
	wakeup_one(data);
	mtx_unlock(&data->msg_mtx);
}

static __inline int
fdata_kick_get(struct fuse_data *data)
{
	DEBUG("0x%x\n", data->dataflag & FDAT_KICK);
	return (data->dataflag & FDAT_KICK);
}

/* <== fuse_data methods */

/* fuse_ticket methods ==> */

/*
 * Tickets are carriers of communication with a fuse daemon.
 * Tickets have a unique id, which should be kept unique
 * among synchronously used tickets. The everyday routine:
 * + Syscall handler gets one by ticket_fetch()
 *     (then, behind the scenes, either a fresh one is allocated
 *     or a cached old one is put into use again)
 * + Sets up data to be passed to the daemon
 *    (there is a message node attached to the ticket)
 * + If she knows that answer will come, sets up the callback node
 *   attached to the ticket, and enqueues it to the callback nodes' list
 * + enqueues the message and kicks fusedev_read()
 * + if daemon gives a proper answer, it will contain the unique id, by which
 *   fusedev_write() can pick the pre-prepared callback node and run it.
 *   There is the standard callback handler which just awakes the syscall
 *   handler who then waits for the response, deal with it by herself.
 *   Another option is to write a full-fledged callback handler which does all
 *   data processing within fusedev_write()'s thread (although we never do
 *   this). Or, if the answer is uninteresting, a NULL handler can be used.
 * + After answer has been consumed, the handler or anyone to whom the control
 *   has been passed supposed to call ticket_drop()
 *     (then, behind the scenes, either the ticket is destroyed, or put into
 *     cache)
 */


static struct fuse_ticket *
ticket_alloc(struct fuse_data *data)
{
	struct fuse_ticket *tick;

	MALLOC(tick, struct fuse_ticket *, sizeof(*tick), M_FUSEMSG,
	       M_WAITOK | M_ZERO);

	tick->unique = data->ticketer++;
	tick->data = data;
	fuse_msgn_init(&tick->msgn, tick, sizeof(struct fuse_in_header));
	fuse_callbn_init(&tick->callbn, tick);
	
	return (tick);
}

static __inline void
ticket_refresh(struct fuse_ticket *tick)
{
		fuse_msgn_refresh(&tick->msgn);
		fuse_callbn_refresh(&tick->callbn);
		tick->miscflag = 0;
		tick->age++;
}

static void
ticket_destroy(struct fuse_ticket *tick)
{
	fuse_msgn_teardown(&tick->msgn);
	fuse_callbn_teardown(&tick->callbn);
	FREE(tick, M_FUSEMSG);
}

__static void
ticket_drop(struct fuse_ticket *tick)
{
	int die = 0;

	/*
	 * Limit below should be tunable by a sysctl?
	 * Probably not worth for the effort.
         */
	if (FUSE_MAX_STORED_FREE_TICKETS &&
	    FUSE_MAX_STORED_FREE_TICKETS <= tick->data->freeticket_counter) {
		die = 1;
		DEBUG("ticket will die\n");
	} else
		ticket_refresh(tick);

	mtx_lock(&tick->data->ticket_mtx);
	ticket_remove_btick(tick);
	if (die) {
		mtx_unlock(&tick->data->ticket_mtx);
		ticket_destroy(tick);
	} else {
		ticket_push_ftick(tick);
		mtx_unlock(&tick->data->ticket_mtx);
	}

	DEBUG("ticket removed\n");				
}

__static __inline void
ticket_invalidate(struct fuse_ticket *tick)
{
	tick->miscflag |= FU_INVAL;
}

static __inline void
ticket_drop_invalid(struct fuse_ticket *tick)
{
	if (tick->miscflag & FU_INVAL)
		ticket_drop(tick);
}


__static struct fuse_ticket *
ticket_fetch(struct fuse_data *data)
{
	struct fuse_ticket *tick;

	DEBUG("fetching ticket\n");
	if (data->freeticket_counter == 0) {
		tick = ticket_alloc(data);
		mtx_lock(&data->ticket_mtx);
	} else {
		mtx_lock(&data->ticket_mtx);
		tick = fdata_pop_ftick(data);
		KASSERT(tick, ("no free ticket available tho counter said there is"));
	}

	ticket_push_btick(tick);
	mtx_unlock(&data->ticket_mtx);

	return (tick);
}

/* <== fuse_ticket methods */


/*********************
 *
 * >>> Other aux routines
 *
 *********************/


__static void
fuse_insert_callback(struct fuse_ticket *tick, fuse_handler_t *handler)
{
	tick->callbn.handler = handler;

	mtx_lock(&tick->data->callback_mtx);
	fuse_callbn_push(&tick->callbn);
	mtx_unlock(&tick->data->callback_mtx);
}

__static void
fuse_insert_message(struct fuse_ticket *tick)
{
	if (tick->miscflag & FU_DIRTY)
		panic("ticket reused without being refreshed");
	tick->miscflag |= FU_DIRTY;
	mtx_lock(&tick->data->msg_mtx);
	fuse_msgn_push(&tick->msgn);
	DEBUG("ring the bell\n");
	wakeup_one(tick->data);
	mtx_unlock(&tick->data->msg_mtx);
}


static __inline int
fuse_ohead_audit(struct fuse_out_header *ohead, struct uio *uio)
{
	fuprintf("\tOut header -- len: %i, error: %i, unique: %llu; iovecs: %d\n",
		ohead->len, ohead->error, (unsigned long long)ohead->unique, uio->uio_iovcnt);

	if (uio->uio_resid + sizeof(struct fuse_out_header) != ohead->len) {
		fuprintf("Format error: body size differs from size claimed by header\n");
		return (EINVAL);
	}

	if (uio->uio_resid && ohead->error) {
		fuprintf("Format error: non zero error but message had a body\n");
		return (EINVAL);
	}

	/* Sanitize the linuxism of negative errnos */
	ohead->error = -(ohead->error);

	return (0);
}	

__static int
fuse_body_audit(struct fuse_callback_node *caliban, size_t blen)
{
	int err = 0;
	enum fuse_opcode opcode;

	opcode = fuse_callbn_get_opcode(caliban);

	DEBUG("op %s (#%d) was sent to us with a body of size %d\n",
 (0 <= opcode && opcode <= fuse_opnames_entries) ? fuse_opnames[opcode] : "???",
 opcode, (int)blen);

	switch (opcode) {
	case FUSE_LOOKUP:
		err = blen == sizeof(struct fuse_entry_out) ? 0 : EINVAL;
		break;
	case FUSE_FORGET:
		panic("a handler has been intalled for FUSE_FORGET");
		break;
	case FUSE_GETATTR:
		err = blen == sizeof(struct fuse_attr_out) ? 0 : EINVAL;
		break;
	case FUSE_SETATTR:
		err = blen == sizeof(struct fuse_attr_out) ? 0 : EINVAL;
		break;
	case FUSE_READLINK:
		err = FUSE_SYMLINK_MAX >= blen ? 0 : EINVAL;
		break;
	case FUSE_SYMLINK:
		err = blen == sizeof(struct fuse_entry_out) ? 0 : EINVAL;
		break;
	case FUSE_MKNOD:
		err = blen == sizeof(struct fuse_entry_out) ? 0 : EINVAL;
		break;
	case FUSE_MKDIR:
		err = blen == sizeof(struct fuse_entry_out) ? 0 : EINVAL;
		break;
	case FUSE_UNLINK:
		err = blen == 0 ? 0 : EINVAL;
		break;
	case FUSE_RMDIR:
		err = blen == 0 ? 0 : EINVAL;
		break;
	case FUSE_RENAME:
		err = blen == 0 ? 0 : EINVAL;
		break;
	case FUSE_LINK:
		err = blen == sizeof(struct fuse_entry_out) ? 0 : EINVAL;
		break;
	case FUSE_OPEN:
		err = blen == sizeof(struct fuse_open_out) ? 0 : EINVAL;
		break;
	case FUSE_READ:
		/*
		 * we do whatever crazyness is needed to read the original
		 * request size, starting from the callback node, and then
		 * check whether available data doesn't grow over that.
		 */
		err = ((struct fuse_read_in *)(
		        (char *)caliban->ticket->msgn.msg.base +
                        sizeof(struct fuse_in_header)
	              ))->size >= blen ? 0 : EINVAL;
		break;
	case FUSE_WRITE:
		err = blen == sizeof(struct fuse_write_out) ? 0 : EINVAL;
		break;
	case FUSE_STATFS:
		err = blen == sizeof(struct fuse_statfs_out) ? 0 : EINVAL;
		break;
	case FUSE_FLUSH:
		err = blen == 0 ? 0 : EINVAL;
		break;
	case FUSE_RELEASE:
		err = blen == 0 ? 0 : EINVAL;
		break;
	case FUSE_FSYNC:
		err = blen == 0 ? 0 : EINVAL;
		break;
	case FUSE_SETXATTR:
		panic("FUSE_SETXATTR implementor has forgotten to define a response body format check");
		break;
	case FUSE_GETXATTR:
		panic("FUSE_GETXATTR implementor has forgotten to define a response body format check");
		break;
	case FUSE_LISTXATTR:
		panic("FUSE_LISTXATTR implementor has forgotten to define a response body format check");
		break;
	case FUSE_REMOVEXATTR:
		panic("FUSE_REMOVEXATTR implementor has forgotten to define a response body format check");
		break;
	case FUSE_INIT:
		err = blen == sizeof(struct fuse_init_in_out) ? 0 : EINVAL;
		break;
	case FUSE_OPENDIR:
		err = blen == sizeof(struct fuse_open_out) ? 0 : EINVAL;
		break;
	case FUSE_READDIR:
		err = ((struct fuse_read_in *)(
		        (char *)caliban->ticket->msgn.msg.base +
                        sizeof(struct fuse_in_header)
	              ))->size >= blen ? 0 : EINVAL;
		break;
	case FUSE_RELEASEDIR:
		err = blen == 0 ? 0 : EINVAL;
		break;
	case FUSE_FSYNCDIR:
		err = blen == 0 ? 0 : EINVAL;
		break;
#ifdef FUSE_GETLK
	case FUSE_GETLK:
		panic("FUSE_GETLK implementor has forgotten to define a response body format check");
		break;
#endif
#ifdef FUSE_SETLK
	case FUSE_SETLK:
		panic("FUSE_SETLK implementor has forgotten to define a response body format check");
		break;
#endif
#ifdef FUSE_SETLKW
	case FUSE_SETLKW:
		panic("FUSE_SETLKW implementor has forgotten to define a response body format check");
		break;
#endif
#ifdef FUSE_ACCESS
	case FUSE_ACCESS:
		panic("FUSE_ACCESS implementor has forgotten to define a response body format check");
		break;
#endif
	default:
		panic("fuse opcodes out of sync");
	}

	if (err)
		DEBUG("body is invalid\n");
	return (err);
}

#define RECTIFY_TDCR(td, cred)			\
do {						\
	if (! (td))				\
		(td) = curthread;		\
	if (! (cred))				\
		(cred) = (td)->td_ucred;	\
} while (0)

static __inline void
fuse_setup_ihead(struct fuse_in_header *ihead, struct fuse_ticket *tick,
                 uint64_t nid, enum fuse_opcode op, size_t blen,
                 struct thread* td, struct ucred *cred)
{
	ihead->len = sizeof(ihead) + blen; /* actually not used by lib */
	ihead->unique = tick->unique;
	ihead->nodeid = nid;
	ihead->opcode = op;
	RECTIFY_TDCR(td, cred);
	ihead->pid = td->td_proc->p_pid;
	ihead->uid = cred->cr_uid;
	ihead->gid = cred->cr_rgid;
	DEBUG2G("in header -- opcode: %2d (\"%s\"), unique: %llu, nodeid: %llu\n",
	      ihead->opcode,
	      fuse_opnames[ihead->opcode],
	      (unsigned long long)ihead->unique,
	      (unsigned long long)ihead->nodeid); 
}

__static __inline struct fuse_data *
fusedev_get_data(struct cdev *fdev)
{
	return (fdev->si_drv1);
}

__static __inline struct sx *
fusedev_get_lock(struct cdev *fdev)
{
	return (&fusedev_get_data(fdev)->slock);
}
	
/********************
 *
 * >>> Callback handlers
 *
 ********************/

/*
 * fuse_standard_handler just pulls in data and wakes up pretender.
 * Doesn't try to interpret data, that's left for the pretender.
 * Though might do a basic size verification before the pull-in takes place
 */

__static int
fuse_standard_handler(struct fuse_callback_node *caliban, struct uio *uio)
{
	int err = 0;
	int dropflag = 0;

	if (caliban->ohead->error)
		goto out;
	if ((err = fuse_body_audit(caliban, uio->uio_resid)))
		goto out;
	if ((err = fuse_callbn_pull_uio(caliban, uio)))
		goto out;

out:
	mtx_lock(&caliban->answer_mtx);
	if (fuse_callbn_answered(caliban))
		/* The requester was interrupted and she set the "answered" flag
		 * to notify us. In this case, we don't have to care about
		 * anything, just drop the ticket and get out as fast as we can.
		 */
		dropflag = 1;
	else {
		fuse_callbn_set_answered(caliban);
		caliban->errno = err;
		wakeup(caliban->ticket);
	}
	mtx_unlock(&caliban->answer_mtx);

	if (dropflag)
		ticket_drop(caliban->ticket);
	return (err);
}

/****************************
 *
 * >>> Fuse device op defs
 *
 ****************************/

/* 
 * Resources are set up on a per-open basis
 */
static int
fusedev_open(struct cdev *dev, int oflags, int devtype, struct thread *td)
{
	struct fuse_data *fdata;	

	/*
	 * We need to ensure data consistency on the main entry points
	 * of the module (so that we won't race with module unloading).
	 * I don't see a better way of doing this than using Giant.
	 * The lightest reliance on Giant is realized by the 
	 * DO_GIANT_MANUALLY == USE_FUSE_LOCK == 1 combination.
	 *
	 * Then the device will operate D_NEEDGIANT-less, and the only
	 * point where Giant is used (apart from the obligatory places
	 * like module (un)load, (un)mount, etc.) is at the beginning
	 * of the device open routine, where destructive racing is
	 * prevented.
	 */

#if DO_GIANT_MANUALLY 
	mtx_lock(&Giant);
#endif
	if (fuse_useco < 0) {
		/* Module unload is going on */
#if DO_GIANT_MANUALLY 
		mtx_unlock(&Giant);
#endif
		DEBUG2G("caught in the middle of unload\n");
		return (ENOENT);
	}
#if DO_GIANT_MANUALLY && USE_FUSE_LOCK
	fuse_useco++;
 	mtx_unlock(&Giant);
#endif

	if (dev->si_usecount > 1)
		goto busy;

	DEBUG2G("device %p\n", dev);

	fdata = fdata_alloc(td->td_ucred);

	FUSE_LOCK;
	if (fusedev_get_data(dev)) {
		FUSE_UNLOCK;
		fdata_destroy(fdata);
		goto busy;
	} else {
#if ! (DO_GIANT_MANUALLY && USE_FUSE_LOCK)
		fuse_useco++;
#endif
		fdata->dataflag |= FDAT_OPENED;
		dev->si_drv1 = fdata;
	}	
	FUSE_UNLOCK;
#if DO_GIANT_MANUALLY && ! USE_FUSE_LOCK
	mtx_unlock(&Giant);
#endif

	DEBUG("Opened device \"fuse\" (that of minor %d) successfully on thread %d.\n", minor(dev), td->td_tid);

	return(0);

busy:
#if DO_GIANT_MANUALLY
#if USE_FUSE_LOCK
	fuse_useco--;
#else
	mtx_unlock(&Giant);
#endif
#endif
	return (EBUSY);
}

static int
fusedev_close(struct cdev *dev, int fflag, int devtype, struct thread *td)
{
	struct fuse_data *data;

#if DO_GIANT_MANUALLY && ! USE_FUSE_LOCK
	mtx_lock(&Giant);
#endif
	FUSE_LOCK;
	data = fusedev_get_data(dev);
	if (! data)
		panic("no fuse data upon fuse device close");
	KASSERT(data->dataflag | FDAT_OPENED,
	        ("fuse device is already closed upon close"));
        data->dataflag &= ~FDAT_OPENED;
	DEBUG2G("mntco %d\n", data->mntco);
	if (data->mntco > 0) {
		FUSE_UNLOCK;
		goto out;
	}
	dev->si_drv1 = NULL;
	FUSE_UNLOCK;

	fdata_destroy(data);

out:
#if DO_GIANT_MANUALLY && ! USE_FUSE_LOCK
	mtx_unlock(&Giant);
#endif
	fuse_useco--;
	
	DEBUG("Closing device \"fuse\" (that of minor %d) on thread %d.\n", minor(dev), td->td_tid);
	return(0);
}

/*
 * fusedev_read hangs on the queue of VFS messages.
 * When it's notified that there is a new one, it picks that and
 * passes up to the daemon
 */
int
fusedev_read(struct cdev *dev, struct uio *uio, int ioflag)
{
	int err = 0;
	struct fuse_data *data;
	struct fuse_msg_node *fmsgn;

	data = fusedev_get_data(dev);

	fuprintf("fuse device being read on thread %d\n", uio->uio_td->td_tid);

	mtx_lock(&data->msg_mtx);
again:
	if (fdata_kick_get(data)) {
		DEBUG("we know early on that reader should be kicked so we don't wait for news\n");
		mtx_unlock(&data->msg_mtx);
		return (ENODEV);
	}

	if ( ! (fmsgn = fdata_pop_msg(data))) {
		err = msleep(data, &data->msg_mtx, PCATCH, "fu msg", 0);
		if (err != 0) {
			mtx_unlock(&data->msg_mtx);
			return (fdata_kick_get(data) ? ENODEV : err);
		}
		fmsgn = fdata_pop_msg(data);
	}
	if (! fmsgn) {
		/*
		 * We can get here if fuse daemon suddenly terminates,
		 * eg, by being hit by a SIGKILL
		 * -- and some other cases, too, tho not totally clear, when
		 * (cv_signal/wakeup_one signals the whole process ?)
		 */
		DEBUG("no message on thread #%d\n", uio->uio_td->td_tid);
		goto again;
	}
	mtx_unlock(&data->msg_mtx);

	if (fdata_kick_get(data)) {
		/*
		 * somebody somewhere -- eg., umount routine --
		 * wants this liaison finished off
		 */
		DEBUG("reader is to be sacked\n");
		if (fmsgn) {
			DEBUG("weird -- \"kick\" is set tho there is message\n");
			ticket_drop_invalid(fmsgn->ticket);
		}
		return (ENODEV); /* This should make the daemon get off of us */
	}
	DEBUG("message got on thread #%d\n", uio->uio_td->td_tid);

	KASSERT(fmsgn->msg_bufdata || fmsgn->msg_bufsize == 0,
	        ("non-null buf pointer with positive size"));
	/*
	 * Why not ban mercilessly stupid daemons who can't keep up with us?
	 * (There is no much use of a partial read here...)
	 */
	if (uio->uio_resid < fmsgn->msg.len) {
		data->dataflag |= FDAT_KICK;
		err = ENODEV;
	}

	if (! err)
		err = uiomove(fmsgn->msg.base,
	                      MIN(uio->uio_resid, fmsgn->msg.len), uio);

	if (! err && fmsgn->msg_bufsize != 0)
		err = uiomove(fmsgn->msg_bufdata,
		              MIN(uio->uio_resid, fmsgn->msg_bufsize), uio);

	/*
	 * fusedev_read will drop "invalidated" tickets
	 * (used when the one who inserted the message thinks the daemon
	 * won't aswer )
	 */
	ticket_drop_invalid(fmsgn->ticket);

	return(err);
}

/*
 * fusedev_write first reads the header sent by the daemon.
 * If that's OK, looks up ticket/callback node by the unique id seen in header.
 * If the callback node contains a handler function, the uio is passed over
 * that.
 */
static int
fusedev_write(struct cdev *dev, struct uio *uio, int ioflag)
{
#if _DEBUG_MSGING
	static int counter=0;
#endif
	struct fuse_out_header *ohead;
	int err = 0;
	struct fuse_data *data;
	struct fuse_callback_node *fcallbn, *fcallbn_temp;
	int found = 0;

	fuprintf("Fuse write -- No: %d, resid: %d, iovcnt: %d, thread: %d\n",
		++counter, uio->uio_resid, uio->uio_iovcnt, uio->uio_td->td_tid);

	if (uio->uio_resid < sizeof(struct fuse_out_header)) {
		fuprintf("\tgot less than a header!\n");
		return (EINVAL);
	}

	MALLOC(ohead, struct fuse_out_header *, sizeof(struct fuse_out_header),
	       M_TEMP, M_WAITOK);

	if ((err = uiomove(ohead, sizeof(struct fuse_out_header), uio)) != 0)
		goto drophead;

	/*	
	 * We check header information (which is redundant) and compare it with
	 * what we see. If we see some inconsistency we discard the whole
	 * answer and proceed on as if it had never existed. In particular, no
	 * pretender will be woken up, regardless the "unique" value in the
	 * header.
	 */
	if ((err = fuse_ohead_audit(ohead, uio)))
		goto drophead;
	
	data = fusedev_get_data(dev);

	/* Pass stuff over to callback if there is one installed */

	/* Looking for ticket with the unique id of header */
	mtx_lock(&data->callback_mtx);
	TAILQ_FOREACH_SAFE(fcallbn, &data->fcallb_head, fcallb_link,
	                   fcallbn_temp) {
		fuprintf("bumped into callback #%llu\n",
                         (unsigned long long)fcallbn->ticket->unique);
		if (fcallbn->ticket->unique == ohead->unique) {
			found = 1;
			fuse_callbn_remove(fcallbn);
			break;
		}
	}
	mtx_unlock(&data->callback_mtx);

	if (found) {
		if (fcallbn->handler) {
			/* We found a callback with proper handler.
			 * In this case the out header will be 0wnd by
			 * the callback, so the fun of freeing that is
			 * left for her.
			 * (Then, by all chance, she'll just get that's done
			 * via ticket_drop(), so no manual mucking around...)
			 */
			fcallbn->ohead = ohead;
			err = fcallbn->handler(fcallbn, uio);
			fuprintf("stuff been passed over to a callback\n");
		} else {
			/* pretender doesn't wanna do anything with answer */
			fuprintf("stuff devalidated, so we drop it\n");
			ticket_drop(fcallbn->ticket);
			goto drophead;
		}
	} else {
		/* no callback at all! */
		/*
		 * this is BAD:
		 *       + if unique id was generated by fuse userspace,
		 *         that's way too bogus. Userspace should only
		 *         pass unique ids back and forth.
		 *       + else we generated the unique id by ticket_fetch()
		 *         somewhere in the foggy past.
		 *         Now that ticket is either
		 *       	+ dropped, which means it could have been
		 *       	  reused by now, and then this answer we work
		 *                 with would be misdelivered (it's
		 *                 just luck that it's not the case now)
		 *       	+ not dropped, which means it's either
		 *       		+ lost: will be kept on busy list for
		 *       		  the whole lifetime
		 *       		+ someone will drop it randomly without
		 *       		  any event based triggering (this
		 *       		  would be the place for such a thing),
		 *       		   which is just nonsense.
		 */
		fuprintf("erhm, no handler for this response\n");
		goto drophead;
	}

	goto out;

drophead:
	NFREE(ohead,M_TEMP);

out:
	return(err);

}


/***********************************************************************
 *
 * >>>  High-level interface to the messaging facilities of fuse devices
 *
 ***********************************************************************/

/* aka stage 1 */

static int
fdisp_get_vopdata(struct fuse_dispatcher *fdip, struct mount *mp)	
{
	fdip->tick = NULL;
	fdip->fdev = ((struct fuse_mnt_data *)mp->mnt_data)->fdev;
	fdip->slock = fusedev_get_lock(fdip->fdev);
	sx_slock(fdip->slock);
	fdip->data = fusedev_get_data(fdip->fdev);
	if (! (fdip->data->dataflag & FDAT_OPENED)) {
		sx_sunlock(fdip->slock);
		return EIO;
	}

	return (0);
}

/* aka stage 2 */

static void
fdisp_prepare_msg(struct fuse_dispatcher *fdip, enum fuse_opcode op,
                  uint64_t nid, struct thread *td, struct ucred *cred)
{
	if (! fdip->tick)	
		fdip->tick = ticket_fetch(fdip->data);

	FUSE_DIMALLOC(&fdip->tick->msgn.msg, fdip->finh,
	              fdip->indata, fdip->iosize);
	fuse_setup_ihead(fdip->finh, fdip->tick, nid, op, fdip->iosize, td, cred);
}

/* prev. two in one */

static __inline int
fdisp_prepare_all(struct fuse_dispatcher *fdip, enum fuse_opcode op,
                  struct vnode *vp, struct thread *td, struct ucred *cred)
{
	int err = 0;

	if ((err = fdisp_get_vopdata(fdip, vp->v_mount)))
		return (err);

	fdisp_prepare_msg(fdip, op, VTOI(vp), td, cred);

	return (0);
}

/* aka stage 3 */

__static int
fdisp_wait_answ(struct fuse_dispatcher *fdip)
{
	int err = 0;
	fdip->answ_stat = 0;

	fuse_insert_callback(fdip->tick, fuse_standard_handler);
	fuse_insert_message(fdip->tick);
	if ((err = fuse_callbn_wait_answer(&fdip->tick->callbn))) {
		/* Uh-huh, we got interrupted... */

#if ! DONT_TRY_HARD_PREVENT_IO_IN_VAIN
		struct fuse_callback_node *fcallbn;
		unsigned age;
#endif

		mtx_lock(&fdip->tick->callbn.answer_mtx);
		if (fuse_callbn_answered(&fdip->tick->callbn)) {
			/*
			 * Just between noticing the interrupt and getting here,
			 * the standard handler has completed his job.
			 * So we drop the ticket and exit as usual.
			 */
			mtx_unlock(&fdip->tick->callbn.answer_mtx);
			goto out;
		} else {
			/*
			 * So we were faster than the standard handler.
			 * Then by setting the answered flag we get *him*
			 * to drop the ticket.
			 */
			age = fdip->tick->age;
			fuse_callbn_set_answered(&fdip->tick->callbn);
			mtx_unlock(&fdip->tick->callbn.answer_mtx);
#if ! DONT_TRY_HARD_PREVENT_IO_IN_VAIN
			/*
			 * If we are willing to pay with one more locking, we
			 * can save on I/O by getting the device write handler
			 * to drop the ticket. That is, if we are fast enough,
			 * the standard handler -- who does the uiomove --
			 * won't even be called. (No guarantee though for
			 * being fast.)
			 */

			DEBUG2G("interrupted, now trying to prevent io in vain (I'll tell you if it succeeds...)\n");
			mtx_lock(&fdip->data->callback_mtx);
			TAILQ_FOREACH(fcallbn, &fdip->data->fcallb_head,
			              fcallb_link) {
				if (fcallbn->ticket == fdip->tick) {
					if (fdip->tick->age == age) {
						DEBUG2G("preventing io in vain succeeded\n");
						fdip->tick->callbn.handler = NULL;
					}
					break;
				}
			}
			mtx_unlock(&fdip->data->callback_mtx);
#endif
			if (fdip->slock)
				sx_sunlock(fdip->slock);
			return (err);
		}
	}

	if (fdip->tick->callbn.errno) {
		/*
		 * Error with messaging, the VFS layer
		 * won't be particularly interested about
		 * its exact nature
		 */
		err = EIO;
		goto out;
	}

	if ((err = fdip->tick->callbn.ohead->error)) {
		/*
		 * This means a "proper" fuse syscall error.
		 * We record this value so the caller will
		 * be able to know it's not a boring messaging
		 * failure, if she wishes so (and if not, she can
		 * just simply propagate the return value of this routine).
 		 * [XXX Maybe a bitflag would do the job too,
		 * if other flags needed, this will be converted thusly.]
		 */
		fdip->answ_stat = err;
		goto out;
	}

	fdip->answ = fuse_callbn_resp(&fdip->tick->callbn)->base;
	fdip->iosize = fuse_callbn_resp(&fdip->tick->callbn)->len;

	return (0);

out:
	ticket_drop(fdip->tick);
	if (fdip->slock)
		sx_sunlock(fdip->slock);
	return (err);
}

/* stage 1, 2, 3 together for the simple case when only a header is sent up */

static __inline int
fdisp_simple_putget(struct fuse_dispatcher *fdip, enum fuse_opcode op,
                    struct vnode *vp, struct thread *td, struct ucred *cred)
{
	int err = 0;

	fdip->iosize = 0;

	if (!(err = fdisp_prepare_all(fdip, op, vp, td, cred)))
		err = fdisp_wait_answ(fdip);

	return (err);
}

static __inline int
fuse_send_init(struct fuse_data *data, struct thread *td)
{
	int err = 0;
	struct fuse_init_in_out *fiio;
	struct fuse_dispatcher fdi;

	fdi.iosize = sizeof(*fiio);
	fdi.slock = NULL;
	fdi.data = data;
	fdi.tick = NULL;

	fdisp_prepare_msg(&fdi, FUSE_INIT, 0, td, NULL);
	fiio = fdi.indata; 		
	fiio->major = FUSE_KERNEL_VERSION;
	fiio->minor = FUSE_KERNEL_MINOR_VERSION;

	if ((err = fdisp_wait_answ(&fdi)))
		return (err);

	fiio = fdi.answ;
	/* XXX is the following check adequate? */
	if (fiio->major < 7) {
		DEBUG2G("userpace version too low\n");
		err = EPROTONOSUPPORT;
	}

	ticket_drop(fdi.tick);
	return (err);
}

static __inline int
fuse_send_forget(struct mount *mp, struct thread *td, struct ucred *cred,
                 uint64_t nodeid, uint64_t nlookup,
                 struct fuse_dispatcher *fdip)
{
	struct fuse_forget_in *ffi;
	int err = 0;

	DEBUG("sending FORGET with %llu lookups\n", (unsigned long long)nlookup);
	fdip->iosize = sizeof(*ffi);

	if (fdip->tick) {
		KASSERT(fdip->slock, ("fdi has ticket but no lock!"));
		ticket_refresh(fdip->tick);
	} else {
		if ((err = fdisp_get_vopdata(fdip, mp)))
			return (err);
	}
	fdisp_prepare_msg(fdip, FUSE_FORGET, nodeid, td, cred);

	ffi = fdip->indata;
	ffi->nlookup = nlookup;

	ticket_invalidate(fdip->tick);		
	fuse_insert_message(fdip->tick);

	/*
	 * I think we can unlock here -- users of this routine won't wanna
	 * go and fetch a new ticket...
	 * let's just save them from dealing with this lock
	 */
	sx_sunlock(fdip->slock);

	return (0);
}

/*8<--8<--8<--8<--8<--8<--8<--8<--8<--8<--8<--8<--8<--8<--8<--8<--8<--8<--
 *  <---- fusedev.c
 *  Possible file boundary in the future!
 *  fuse_vfs.c ---->
 */

typedef	int fuse_buffeater_t(struct uio *uio, size_t reqsize, void *buf, size_t bufsize, void *param);
typedef int fuse_metrics_t(struct vnode *vp, struct thread *td, struct ucred *cred, int mode, struct fuse_filehandle *fufh, void *param);

static __inline void fuse_vnode_init(struct vnode *vp, struct fuse_vnode_data *fvdat, uint64_t nodeid, enum vtype vtyp);
static vfs_mount_t fuse_mount;
static vfs_unmount_t fuse_unmount;
static vfs_root_t fuse_root;
static vfs_statfs_t fuse_statfs;
static vfs_hash_cmp_t fuse_vnode_cmp;
/* static vfs_vget_t fuse_vget; */
static int fuse_vget_i(struct mount *mp, struct thread *td, uint64_t nodeid, enum vtype vtyp, struct vnode **vpp);
static __inline void fat2vat(struct mount *mp, struct fuse_attr *fat, struct vattr *vap);
static vop_getattr_t fuse_getattr;
static int fuse_recyc_backend(struct vnode *vp, struct thread *td);
static fuse_metrics_t release_filehandle;
static void fuse_filehandle_gc(struct vnode *vp, struct thread *td, struct ucred *cred);
static vop_reclaim_t fuse_reclaim;
static vop_inactive_t fuse_inactive;
static __inline void fuse_vnode_kick(struct vnode *vp, struct thread *td);
static vop_access_t fuse_access;
static __inline int fuse_access_by_attr(struct mount *mp, struct vattr *vap, mode_t mode, struct ucred *cred);
/* static vop_cachedlookup_t fuse_lookup; */
static vop_lookup_t fuse_lookup;
/* static vop_pathconf_t fuse_pathconf; */
static int iterate_filehandles(struct vnode *vp, struct thread *td, struct ucred *cred, int mode, fuse_metrics_t fmetr, void *param);
static int fuse_standard_metrics(struct vnode *vp, struct thread *td, struct ucred *cred, int mode, struct fuse_filehandle *fufh, void *param);
static struct fuse_filehandle *get_filehandle(struct vnode *vp, struct thread *td, struct ucred *cred, int mode);
static vop_open_t fuse_open;
static int fuse_send_release(struct vnode *vp, struct thread *td, struct ucred *cred, struct fuse_filehandle *fufh, int flags);
static fo_close_t fuse_close_f;
static int fuse_read_directbackend(struct vnode *vp, struct fuse_filehandle *fufh, struct uio *uio, struct ucred *cred, struct thread *td, enum fuse_opcode op, fuse_buffeater_t *buffe, void *param);
static int fuse_read_biobackend(struct vnode *vp, struct fuse_filehandle *fufh, struct uio *uio, struct ucred *cred, struct thread *td, enum fuse_opcode op, fuse_buffeater_t *buffe, void *param);
static fo_rdwr_t fuse_read_f; 
static vop_readdir_t fuse_readdir;
static vop_readlink_t fuse_readlink;
static int fuse_newentry_backend(struct vnode *dvp, struct vnode **vpp, struct componentname *cnp, enum fuse_opcode op, void *buf, size_t bufsize, enum vtype vtyp);
static __inline int fuse_make_entry_req(struct vnode *dvp, struct componentname *cnp, enum fuse_opcode op, void *buf, size_t bufsize, struct fuse_dispatcher *fdip);
static __inline int fuse_newentry_core(struct mount *mp, struct vnode **vpp, enum vtype vyp, struct fuse_dispatcher *fdip);
static vop_mknod_t fuse_mknod;
static vop_create_t fuse_create;
static vop_mkdir_t fuse_mkdir;
static vop_link_t fuse_link;
static vop_symlink_t fuse_symlink;
static int fuse_remove_backend(struct vnode *dvp, struct componentname *cnp, enum fuse_opcode op);
static vop_remove_t fuse_remove;
static vop_rmdir_t fuse_rmdir;
static __inline int fuse_rename_core(struct vnode *fdvp, struct componentname *fcnp, struct vnode *tdvp, struct componentname *tcnp);
static vop_rename_t fuse_rename;
static vop_setattr_t fuse_setattr;
static fuse_metrics_t fuse_fsync_filehandle;
static vop_fsync_t fuse_fsync;
static int fuse_write_directbackend(struct vnode *vp, uint64_t fh_id, struct uio *uio, struct ucred *cred, struct thread *td);
static int fuse_write_biobackend(struct vnode *vp, struct uio *uio, struct ucred *cred, struct thread *td);
static fo_rdwr_t fuse_write_f;
static int fuse_strategy_i(struct vnode *vp, struct buf *bp, struct fuse_filehandle *fufh);
static vop_strategy_t fuse_strategy;
static vop_bmap_t fuse_bmap;
static vop_print_t fuse_print;

static b_strategy_t fuse_bufstrategy;

static struct vfsops fuse_vfsops = {
	.vfs_mount   = fuse_mount,
	.vfs_unmount = fuse_unmount,
	.vfs_root    = fuse_root,
	.vfs_statfs  = fuse_statfs,
	/* .vfs_vget    = fuse_vget,
	   .vfs_sync    = vfs_stdsync, */
};

static struct vop_vector fuse_vnops = {
	.vop_default       = &default_vnodeops,
	.vop_inactive      = fuse_inactive,
	.vop_reclaim       = fuse_reclaim,
	.vop_getattr       = fuse_getattr,
	.vop_access        = fuse_access,
	.vop_lookup        = /* vfs_cache_lookup,
	.vop_cachedlookup  = */ fuse_lookup,
	//.vop_pathconf      = fuse_pathconf,
	.vop_open          = fuse_open,
	.vop_readdir       = fuse_readdir,
	.vop_readlink      = fuse_readlink,
	.vop_mknod         = fuse_mknod,
	.vop_create        = fuse_create,
	.vop_mkdir         = fuse_mkdir,
	.vop_link          = fuse_link,
	.vop_symlink       = fuse_symlink,
	.vop_remove        = fuse_remove,
	.vop_rmdir         = fuse_rmdir,
	.vop_rename        = fuse_rename,
	.vop_setattr       = fuse_setattr,
	.vop_fsync         = fuse_fsync,
	.vop_strategy      = fuse_strategy,
	.vop_bmap          = fuse_bmap,
	.vop_print         = fuse_print,
};

static struct fileops fuse_fileops;

static struct buf_ops fuse_bufops = {
	.bop_name       = "Fuse",
	.bop_strategy   = fuse_bufstrategy,
	.bop_write      = bufwrite,
	.bop_sync       = bufsync,
};

MALLOC_DEFINE(M_FUSEFS, "fuse filesystem", "buffer for fuse vfs layer");
MALLOC_DEFINE(M_FUSEFH, "fuse filehandles", "buffer for fuse filehandles");

static fuse_buffeater_t fuse_std_buffeater; 
static fuse_buffeater_t fuse_dir_buffeater; 


/* 
 * As I don't use the name cache, I can muck with the v_dd field.
 * (First I tried to use a private field but v_dd has to be set
 * as vn_fullpath1() of vfs_cache.c b0rks without it, and __getcwd()
 * relies on that function
*/
#define GETPARENT(pvp, vp)						\
do {									\
	KASSERT((vp)->v_dd != (vp) || (vp)->v_vflag & VV_ROOT,	\
	        ("self-parented non-root"));				\
  	(pvp) = (vp)->v_dd;						\
} while (0)

#define SETPARENT(vp, pvp)						\
do {									\
	KASSERT((vp) != (pvp) || (vp)->v_vflag & VV_ROOT,		\
                ("attempt to make non-root node parented by itself"));	\
	(vp)->v_dd = (pvp);						\
} while (0)

/********
 *
 * >>> More aux routines
 *
 *******/

static __inline void
fuse_vnode_init(struct vnode *vp, struct fuse_vnode_data *fvdat,
	        uint64_t nodeid, enum vtype vtyp)
{
	fvdat->nid = nodeid;
	vp->v_data = fvdat;
	SETPARENT(vp, (vp->v_vflag & VV_ROOT) ? vp : NULL);
	vp->v_type = vtyp;

	sx_init(&fvdat->fh_lock, "lock for fuse filehandles");
	LIST_INIT(&fvdat->fh_head);

	vp->v_bufobj.bo_ops = &fuse_bufops;
	vp->v_bufobj.bo_private = vp;
}	

static __inline void
fuse_vnode_kick(struct vnode *vp, struct thread *td)
{
	struct fuse_vnode_data *fvdat;

	if (! td)
		td = curthread;

	if (vp->v_vflag & VV_ROOT) {
		fdata_kick_set(fusedev_get_data(((struct fuse_mnt_data *)vp->v_mount->mnt_data)->fdev));
		return;
	}
	fvdat = vp->v_data;
	fvdat->nlookup = 0;
	DEBUG("pfft...\n"); 
	fuse_filehandle_gc(vp, td, NULL);		
	vnode_destroy_vobject(vp);
	/*
	 * this implies we won't get feedback on recycling
	 * (other than panicking, or the lack of that)
	 * but creating a customized set of bad vnode ops
	 * would be too much hassle...
	 */
	vp->v_op = &dead_vnodeops;
	fuse_recyc_backend(vp, td);
	vput(vp);
}

/*************
 *
 * >>> VFS ops
 *
 *************/

/*
 * Mount system call
 */
static int
fuse_mount(struct mount *mp, struct thread *td)
{
	int err = 0;
	int typelen = 0;
	size_t len;
	char *fspec;
	struct vnode *devvp;
	struct vfsoptlist *opts;
	struct nameidata nd, *ndp = &nd;
	struct cdev *fdev;
	struct sx *slock;
	struct fuse_data *data;
	struct fuse_mnt_data *fmnt;
	struct vnode *rvp;
	struct fuse_vnode_data *fvdat;

	GIANT_REQUIRED;
	KASSERT(fuse_useco >= 0,
	        ("negative fuse usecount despite Giant"));

	if (mp->mnt_flag & MNT_UPDATE) {
		uprintf("fuse: updating mounts is not supported\n");
		return (EOPNOTSUPP);
	}

	/*
	 * there are some hooks for checking MNT_RDONLY, but to say
	 * it's supported, ie. all writing is blocked, a thoroughout
	 * code review would be necessary.
	 */
	if (mp->mnt_flag & MNT_RDONLY) {
		uprintf("fuse: mounting readonly is not supported\n");
		return (EOPNOTSUPP);
	}

	mp->mnt_flag |= MNT_SYNCHRONOUS; 
	/* Get the new options passed to mount */
	opts = mp->mnt_optnew;

	/* `fspath' contains the mount point (eg. /mnt/fuse/sshfs); REQUIRED */
	if ((err = vfs_getopt(opts, "fspath", NULL, NULL)))
		return (EINVAL);

	/* `from' contains the device name (eg. /dev/fuse0); REQUIRED */
	fspec = NULL;
	if ((err = vfs_getopt(opts, "from", (void **)&fspec, &typelen)))
		return (err);
	if (!fspec || fspec[typelen - 1] != '\0')
		return (EINVAL);

	mp->mnt_data = NULL;

	/*
	 * Not an update, or updating the name: look up the name
	 * and verify that it refers to a sensible disk device.
	 */ 

	NDINIT(ndp, LOOKUP, FOLLOW, UIO_SYSSPACE, fspec, td);
	if ((err = namei(ndp)) != 0)
		return (err);
	NDFREE(ndp, NDF_ONLY_PNBUF);
	devvp = ndp->ni_vp;

	if (devvp->v_type != VCHR) {
		vrele(devvp);
		return (ENXIO);
	}
	
	fdev = devvp->v_rdev;
	dev_ref(fdev);
	/*
	 * according to coda code, no extra lock is needed --
	 * although in sys/vnode.h this field is marked "v"
	 */
	vrele(devvp);
	
	FUSE_LOCK;
	if (! fdev->si_devsw ||
	    strcmp("fuse", fdev->si_devsw->d_name) ||
	    ! (slock = fusedev_get_lock(fdev))) {
		FUSE_UNLOCK;
		return (ENXIO);
	}

	if ((data = fusedev_get_data(fdev)) &&
	    data->dataflag & FDAT_OPENED)
		data->mntco++;
	else {
		FUSE_UNLOCK;
		dev_rel(fdev);
		return (ENXIO);
	}	
	FUSE_UNLOCK;

	MALLOC(fmnt, struct fuse_mnt_data *, sizeof(*fmnt), M_FUSEFS,
	       M_WAITOK| M_ZERO);
	
	fmnt->fdev = fdev;
	fmnt->mp = mp;

	vfs_flagopt(opts, "private", &fmnt->mntopts, FUSEFS_PRIVATE); 
	vfs_flagopt(opts, "neglect_shares", &fmnt->mntopts,
	            FUSEFS_NEGLECT_SHARES); 
	vfs_flagopt(opts, "allow_other", &fmnt->mntopts, FUSEFS_DAEMON_CAN_SPY); 
	vfs_flagopt(opts, "push_symlinks_in", &fmnt->mntopts, FUSEFS_PUSH_SYMLINKS_IN);

	if (fdata_kick_get(data))
		err = ENOTCONN;
	if (fmnt->mntopts & FUSEFS_DAEMON_CAN_SPY && suser(td)) {
		uprintf("only root can use \"allow_other\"\n");
		err = EPERM;
	}

	if (err)
		goto out;


	DEBUG2G("mntopts 0x%x\n", fmnt->mntopts);

	/* Sanity + permission checks */ 

	if (! data->daemoncred)
		panic("fuse daemon found, but identity unknown");

	sx_xlock(slock);
	if (data->mp) {
		fmnt->master = data->mp->mnt_data;
		fmnt->mntopts |= FUSEFS_SECONDARY;
		if (fmnt->master->mntopts & FUSEFS_BUSY)
			/*
			 * Umount attempt is going on
			 */
			err = EBUSY;
		if (fmnt->master->mntopts & FUSEFS_PRIVATE)
			/*
			 * device is owned and owner doesn't
			 * wanna share it with us 
			 */
			err = EPERM;
		if (fmnt->mntopts & ~FUSEFS_SECONDARY)
			/*
			 * Secondary mounts not allowed to have
			 * options (basicly, that would be
			 * useless though harmless, just let's
			 * be explicit about it)
			 */
			err = EINVAL;
	} else {
		if (suser(td) &&
	            td->td_ucred->cr_uid != data->daemoncred->cr_uid)
			/* we are not allowed to do the first mount */
			err = EPERM;
	}

	if (err) {
		sx_xunlock(slock);
		goto out;
	}

	if (fmnt->mntopts & FUSEFS_SECONDARY)
		LIST_INSERT_HEAD(&fmnt->master->slaves_head, fmnt, slaves_link);
	else {
		LIST_INIT(&fmnt->slaves_head);

		/* Now handshaking with daemon */

		if ((err = fuse_send_init(data, td))) {
			/*
			 * Something went wrong with sending init --
			 * eg., bad version number
			 * Thus we tell the daemon it's time to get off
			 */
			fdata_kick_set(data);
	
			sx_xunlock(slock);
			goto out;
		}
	}

	/* We need this here as this slot is used by getnewvnode() */
	mp->mnt_stat.f_iosize = PAGE_SIZE;

	mp->mnt_data = fmnt;

	/* code stolen from portalfs */

	if (fmnt->mntopts & FUSEFS_SECONDARY)
		goto rootdone;

	MALLOC(fvdat, struct fuse_vnode_data *, sizeof(*fvdat), M_FUSEFS,
	       M_WAITOK | M_ZERO);

	err = getnewvnode("fuse", mp, &fuse_vnops, &rvp);

	if (err) {
		fdata_kick_set(data);
		sx_xunlock(slock);
		FREE(fvdat, M_FUSEFS);
	        goto out;
	}

	/*
	 * FUSE_ROOT_INODE as an inode number will be resolved directly.
	 * without resorting to the vfs hashing mechanism, thus it also
	 * can be inserted directly to the v_hash slot.
	 */
	rvp->v_hash = FUSE_ROOT_INODE;
	fmnt->rvp = rvp; 	
	fuse_vnode_init(rvp, fvdat, FUSE_ROOT_INODE, VDIR);
	rvp->v_vflag |= VV_ROOT;

rootdone:

	if (! (fmnt->mntopts & FUSEFS_SECONDARY)) {
		data->mp = mp;
#if ! REALTIME_TRACK_UNPRIVPROCDBG
		fmnt->mntopts &= ~FUSEFS_UNPRIVPROCDBG;
		fmnt->mntopts |= get_unprivileged_proc_debug(td) ? FUSEFS_UNPRIVPROCDBG : 0;
#endif
	}

	sx_xunlock(slock);

	vfs_getnewfsid(mp);	
	mp->mnt_flag |= MNT_LOCAL;

	len = typelen;
	copystr(fspec, mp->mnt_stat.f_mntfromname, MNAMELEN - 1, &len);
	if (fmnt->mntopts & FUSEFS_SECONDARY && len >= 1) {
		/*
		 * I've considered using s1, s2,... for shares, instead
		 * #1, #2,... as s* is more conventional...
		 * Finally I chose #* as I feel it's more expressive
		 * and s* can be misleading (slices are strung up
		 * serially, shares exist in parallell)
		 */
		mp->mnt_stat.f_mntfromname[len-1] = '#';
		if (len < MNAMELEN - 1) {
			/* Duhh, ain't there a better way of doing this? */
			int log = 0, lim = 1;
			while (data->mntco > lim) {
				lim *= 10;
				log++;
			}
			if (log >= MNAMELEN - len) {
				mp->mnt_stat.f_mntfromname[len] = '?';
				len++;
			} else {
				sprintf(mp->mnt_stat.f_mntfromname + len, "%d",
				        data->mntco - 1);
				len += log;
			}
		}
	}
	bzero(mp->mnt_stat.f_mntfromname + len, MNAMELEN - len);
	DEBUG2G("mp %p: %s\n", mp, mp->mnt_stat.f_mntfromname);

out:
	if (err ) {
		data->mntco--;
		FUSE_LOCK;
		if (data->mntco == 0 && ! (data->dataflag & FDAT_OPENED)) {
			fdev->si_drv1 = NULL;
			fdata_destroy(data);
		}
		FUSE_UNLOCK;
		dev_rel(fdev);
		FREE(fmnt, M_FUSEFS);
	}
	return (err);
}

/*
 * Unmount system call
 */
static int
fuse_unmount(struct mount *mp, int mntflags, struct thread *td)
{
	int flags = 0, err = 0;
	struct sx *slock;
	struct fuse_data *data;
	struct fuse_mnt_data *fmnt;

	GIANT_REQUIRED;

	DEBUG2G("mp %p: %s\n", mp, mp->mnt_stat.f_mntfromname);
	/* Flag handling */
	if (mntflags & MNT_FORCE)
		flags |= FORCECLOSE;

	fmnt = mp->mnt_data;
	slock = fusedev_get_lock(fmnt->fdev);

	if (! (fmnt->mntopts & FUSEFS_SECONDARY)) {
#if _DEBUG		
		struct vnode *vp, *nvp;
#endif

		sx_slock(slock);
		if (! (mntflags & MNT_FORCE ||
		    fmnt->mntopts & FUSEFS_NEGLECT_SHARES ||
		    LIST_EMPTY(&fmnt->slaves_head))) {
			sx_sunlock(slock);
			return (EBUSY);
		}
		/* setting flag protecting lock upgrade */
		fmnt->mntopts |= FUSEFS_BUSY;
		sx_sunlock(slock);
#if _DEBUG		
		MNT_ILOCK(mp);
		MNT_VNODE_FOREACH(vp, mp, nvp) {
			DEBUG2G("\n");
			vn_printf(vp, "...");
		}
		MNT_IUNLOCK(mp);
#endif

		/* Flush files -> vflush */
		/* There is 1 extra root vnode reference (mp->mnt_data). */
		if ((err = vflush(mp, 1, flags, td))) {
			DEBUG2G("err %d\n", err);
			fmnt->mntopts &= ~FUSEFS_BUSY;
			return (err);
		}
	}

	data = fusedev_get_data(fmnt->fdev);
	sx_xlock(slock);
	if (fmnt->mntopts & FUSEFS_SECONDARY) {
		if (fmnt->master)
			LIST_REMOVE(fmnt, slaves_link);
	} else {
		struct fuse_mnt_data *x_fmnt;

		fmnt->mntopts &= ~FUSEFS_BUSY;

		LIST_FOREACH(x_fmnt, &fmnt->slaves_head, slaves_link)
			x_fmnt->master = NULL;
		if ((data = fusedev_get_data(fmnt->fdev)))
			fdata_kick_set(data);

		data->mp = NULL;
	}
	sx_xunlock(slock);

	data->mntco--;
	FUSE_LOCK;
	DEBUG2G("mntco %d, opened 0x%x\n",
	        data->mntco, data->dataflag & FDAT_OPENED);
	if (data->mntco == 0 && ! (data->dataflag & FDAT_OPENED)) {
		fmnt->fdev->si_drv1 = NULL;
		fdata_destroy(data);
	}
	FUSE_UNLOCK;
	dev_rel(fmnt->fdev);
	mp->mnt_data = NULL;
	FREE(fmnt, M_FUSEFS);

	/* Other guys do this, I don't know what it is good for... */
	mp->mnt_flag &= ~MNT_LOCAL;

	return (0);
}		

/* stolen from portalfs */
static int
fuse_root(struct mount *mp, int flags, struct vnode **vpp, struct thread *td)
{
	/*
	 * Return locked reference to root.
	 */
	struct fuse_mnt_data *fmnt = mp->mnt_data;
	struct vnode *vp;

	DEBUG2G("mp %p: %s\n", mp, mp->mnt_stat.f_mntfromname);

	if (fmnt->mntopts & FUSEFS_SECONDARY) {
		struct sx *slock;
		int err;

		slock = fusedev_get_lock(fmnt->fdev);
		sx_slock(slock);
		if (fmnt->master)
			err = fuse_root(fmnt->master->mp, flags, vpp, td);
		else
			err = ENXIO;
		sx_sunlock(slock);
		return (err);
	}

	vp = fmnt->rvp;
	vref(vp);
	vn_lock(vp, flags | LK_RETRY, td);
	*vpp = vp;
#if _DEBUG2G
	DEBUG2G("root node:\n");
	vn_printf(vp, " * ");
#endif
	return (0);
}

static int
fuse_statfs(struct mount *mp, struct statfs *sbp, struct thread *td)
{
	struct fuse_dispatcher fdi;
	struct fuse_statfs_out *fsfo;
	struct fuse_mnt_data *fmnt;
	int err = 0;	

	DEBUG2G("mp %p: %s\n", mp, mp->mnt_stat.f_mntfromname);
	fmnt = mp->mnt_data;

	if (fmnt->mntopts & FUSEFS_SECONDARY) {
		struct sx *slock;

		slock = fusedev_get_lock(fmnt->fdev);
		sx_slock(slock);
		if (fmnt->master)
			err = fuse_statfs(fmnt->master->mp, sbp, td);
		else
			err = ENXIO;
		sx_sunlock(slock);
		return (err);
	}

	if ((err = fdisp_simple_putget(&fdi, FUSE_STATFS, fmnt->rvp, td, NULL)))
		return (err);

	fsfo = fdi.answ;

	sbp->f_blocks  = fsfo->st.blocks;
	sbp->f_bfree   = fsfo->st.bfree;
	sbp->f_bavail  = fsfo->st.bavail;
	sbp->f_files   = fsfo->st.files;
	sbp->f_ffree   = fsfo->st.ffree; /* cast from uint64_t to int64_t */
	sbp->f_bsize   = fsfo->st.bsize; /* cast from uint32_t to int64_t */
	sbp->f_namemax = fsfo->st.namelen;	

	DEBUG("fuse_statfs_out -- blocks: %llu, bfree: %llu, bavail: %llu, files: %llu, ffree: %llu, bsize: %i, namelen: %i\n",
	(unsigned long long)fsfo->st.blocks, (unsigned long long)fsfo->st.bfree, (unsigned long long)fsfo->st.bavail, (unsigned long long)fsfo->st.files, (unsigned long long)fsfo->st.ffree, fsfo->st.bsize, fsfo->st.namelen);

	ticket_drop(fdi.tick);
	sx_sunlock(fdi.slock);

	return (0);
}

static int
fuse_vnode_cmp(struct vnode *vp, void *nidp)
{
	return (((struct fuse_vnode_data *)vp->v_data)->nid != *((uint64_t *)nidp));
}

/* ..._i, as "internal" */

static int
fuse_vget_i(struct mount *mp, struct thread *td, uint64_t nodeid,
	    enum vtype vtyp, struct vnode **vpp)
{
	int err = 0;
	struct fuse_mnt_data *fmnt;
	struct fuse_vnode_data *fvdat;
	struct vnode *vp2;
	int myflags = LK_EXCLUSIVE;

	DEBUG2G("mp %p: %s\n", mp, mp->mnt_stat.f_mntfromname);
	DEBUG("been asked for vno #%llu\n", (unsigned long long)nodeid);

	fmnt = mp->mnt_data;
	if (nodeid == FUSE_ROOT_INODE) {
		vpp = &fmnt->rvp;
		vn_lock(*vpp, myflags | LK_RETRY, td);
		return (0);
	}

	DEBUG2G("mp %p: %s\n", mp, mp->mnt_stat.f_mntfromname);

	if ((err = vfs_hash_get(mp, nodeid, /*flags*/ myflags, td, vpp, fuse_vnode_cmp, &nodeid)))
		return (err);

audit:
	if (*vpp) {
		DEBUG("vnode taken from hash\n");
		if ((*vpp)->v_type == vtyp) {
			return (0);
		} else
			fuse_vnode_kick(*vpp, td);
	}

	/* as the big guys say, malloc for your data before getnewvnode() */

	MALLOC(fvdat, struct fuse_vnode_data *, sizeof(*fvdat), M_FUSEFS,
	       M_WAITOK | M_ZERO);
	err = getnewvnode("fuse", mp, &fuse_vnops, vpp);
#if _DEBUG
	DEBUG2G("allocated new vnode:\n");
	vn_printf(*vpp, " * ");
#endif

	if (err) {
	        FREE(fvdat, M_FUSEFS);
		VNASSERT(! *vpp, *vpp,
		         ("Maybe I misunderstood how getnewvnode() works..."));
	        return (err);
	}

	err = vfs_hash_insert(*vpp, nodeid, /*flags*/ myflags, td, &vp2, fuse_vnode_cmp, &nodeid);

	if (err) {
		FREE(fvdat, M_FUSEFS);
		/* XXX do something to tear down allocated inode?
		 * (other fss' code doesn't, so neither I...)
		 */ 
		return (err);
	}

	if (vp2) {
		DEBUG("gee, vfs hash collision for node #%llu\n", (unsigned long long)nodeid);
		vpp = &vp2;
		goto audit;
	}

	fuse_vnode_init(*vpp, fvdat, nodeid, vtyp);
#if _DEBUG
	DEBUG2G("\n");	
	vn_printf(*vpp, " * ");
#endif
	return (err);
}



/******************
 *
 * >>> Vnode ops
 *
 ******************/




static int
fuse_recyc_backend(struct vnode *vp, struct thread *td)
{
	struct fuse_vnode_data *fvdat;

	if (! vp->v_data) {
		DEBUG("got a dataless node\n");
		return (0);
	}

	DEBUG("getting at vnode of ino %llu\n", VTOI32(vp));

	fvdat = vp->v_data;

	if (! (vp->v_vflag & VV_ROOT)) {
		vfs_hash_remove(vp);

		if (fvdat && fvdat->nlookup) {
			struct fuse_dispatcher fdi;

			fdi.tick = NULL;
			/*
			 * sending a FORGET is just courtesy toward the daemon
			 * (who might as well be as dead as a stone at this
			 * point), so we don't care a damned shit about it if
			 * something woes during this action.
			 */
			fuse_send_forget(vp->v_mount, td, NULL, VTOI(vp),
			                 fvdat->nlookup, &fdi);
		}
	}
	vp->v_data = NULL;

	/*
	 * Taking down fuse_vnode_data structures is just hooked in here...
	 * no separate destructor.
	 */
	if (fvdat) {
		sx_destroy(&fvdat->fh_lock);
		FREE(fvdat, M_FUSEFS);
	}

	/* XXX what does it mean exactly to return a non-zero value from here? */
	return (0);
}

static int
release_filehandle(struct vnode *vp, struct thread *td, struct ucred *cred,
		   int mode, struct fuse_filehandle *fufh, void *param)
{
	KASSERT(fufh->useco >= 0, ("negative use count for fuse filehandle"));
	KASSERT(! fufh->fp || fufh->useco > 0,
	        ("filehandle bound with 0 use counter"));
	DEBUG2G("vnode #%llu, fufh owner %p, useco %d\n", VTOI32(vp), fufh->fp, fufh->useco);
	if (! fufh->fp && fufh->useco == 0) {
		LIST_REMOVE(fufh, fh_link);
		fuse_send_release(vp, td, cred, fufh, fufh->mode);
	}

	return (0);		
}

static void
fuse_filehandle_gc(struct vnode *vp, struct thread *td, struct ucred *cred)
{
	struct fuse_vnode_data *fvdat = vp->v_data;

	sx_xlock(&fvdat->fh_lock);
	iterate_filehandles(vp, td, cred, 0, release_filehandle, NULL);
	sx_xunlock(&fvdat->fh_lock);
}

static int
fuse_reclaim(struct vop_reclaim_args *ap)
{
	struct vnode *vp = ap->a_vp;	
	struct thread *td = ap->a_td;

	DEBUG("pfft...\n");
#if _DEBUG
	DEBUG2G("=============>\n");
	kdb_backtrace();
	vn_printf(vp, " ");
	DEBUG2G("<=============\n");
#endif
	fuse_filehandle_gc(vp, td, NULL);		
	vnode_destroy_vobject(vp);
	return (fuse_recyc_backend(vp, td));
}

static int
fuse_inactive(struct vop_inactive_args *ap)
{
	struct vnode *vp = ap->a_vp;	
	struct thread *td = ap->a_td;

	int err;

	DEBUG("getting at vnode of ino %llu\n", VTOI32(vp));
#if _DEBUG
	DEBUG2G("=============>\n");
	kdb_backtrace();
	vn_printf(vp, " ");
	DEBUG2G("<=============\n");
#endif

	fuse_filehandle_gc(vp, td, NULL);		

#if ! DONT_STORE_FS_MAP
	/*
	 * Dropping vnodes when they are not in use would mean that
	 * the respective inode must be freed on the daemon's side
	 * (this is ia necessity according to the protocol: if
	 * we drop the vnode, then we lose the associated nlookup
	 * value, which is required to be kept during the lifetime of
	 * the inode, as at the end of the inode's lifetime this info
	 * is required for properly killing it (cf. forget_node() of 
	 * lib/fuse.c and the assertion in it), that is, the inode
	 * can't be kept alive further than the vnode).
	 *
	 * So either we aggressively kill both kind of beasts and don't
	 * store permanently a "map" of the fs (thus, eg, same files
	 * pop up with different inode numbers upon different lookups, 
	 * which is weird), or we keep them both and don't do anything when 
	 * the vnode usecount falls to zero (ie. use a peaceful vop_inactive
	 * which throws away only rotten vegetables).
	 *
	 * The latter approach seems to be the better, partly because of
	 * the above mentioned weirdness, but mostly because fs ops done
	 * by the daemon are considered expensive.
	 * 
	 * [To add, all this reasoning doesn't imply I understand the 
	 * reason for the enforcement of the thigt synchronization
	 * between in-kernel vnodes and userspace inodes.]
	 */
	if (vp->v_data)
		return (0);
#endif

	if ((err = fuse_recyc_backend(vp, td)))
		return err;

	vrecycle(vp, td);

	return (0);
}

/*
 *  It's by-and-large reversing vn_stat() of kern/vfs_vnops.c
 */

static __inline void
fat2vat(struct mount *mp, struct fuse_attr *fat, struct vattr *vap)
{
	DEBUG("node #%llu, mode 0%o\n", (unsigned long long)fat->ino, fat->mode);

	vap->va_fsid = mp->mnt_stat.f_fsid.val[0];
	vap->va_fileid = fat->ino; /* XXX cast from 64 bits to 32 */
	vap->va_mode = fat->mode & ~S_IFMT;
	
	vap->va_nlink     = fat->nlink;
	vap->va_uid       = fat->uid;
	vap->va_gid       = fat->gid;
	vap->va_rdev      = fat->rdev;
	vap->va_size      = fat->size;
	vap->va_atime.tv_sec  = fat->atime; /* XXX on some platforms cast from 64 bits to 32 */
	vap->va_atime.tv_nsec = fat->atimensec;
	vap->va_mtime.tv_sec  = fat->mtime;
	vap->va_mtime.tv_nsec = fat->mtimensec;
	vap->va_ctime.tv_sec  = fat->ctime;
	vap->va_ctime.tv_nsec = fat->ctimensec;
	vap->va_blocksize = PAGE_SIZE;
	vap->va_type = IFTOVT(fat->mode);

#if (S_BLKSIZE == 512)
	/* Optimize this case */
	vap->va_bytes = fat->blocks << 9;
#else
	vap->va_bytes = fat->blocks * S_BLKSIZE;
#endif
}

static int
fuse_getattr(struct vop_getattr_args *ap)
{
	struct vnode *vp = ap->a_vp;
	struct vattr *vap = ap->a_vap;
	struct ucred *cred = ap->a_cred;
	struct thread *td = ap->a_td;

	struct fuse_mnt_data *fmnt = vp->v_mount->mnt_data;
	struct fuse_dispatcher fdi;
	int err = 0;

	if ((err = fdisp_simple_putget(&fdi, FUSE_GETATTR, vp, td, cred)))
		return (err);

	/*
	 * XXX we discard timeout related info,
 	 * just go for the attributes
 	 */ 
	fat2vat(vp->v_mount, &((struct fuse_attr_out *)fdi.answ)->attr, vap);
	
	ticket_drop(fdi.tick);

	if (! (fmnt->mntopts & FUSEFS_DAEMON_CAN_SPY)) {
		/*
		 * The policy is to forbid a user from using the filesystem,
		 * if she is *more* privileged than the daemon. This is to
		 * protect the power user from the daemon spying on her I/O
		 * operations.
		 *
		 * If a user wouldn't mind this, she can relax this check
		 * by doing a shared mount.
		 */

		int denied;

		if ((denied = cr_candebug(
#if REALTIME_TRACK_UNPRIVPROCDBG
		                get_unprivileged_proc_debug(td),
#else
				fmnt->mntopts & FUSEFS_UNPRIVPROCDBG, 
#endif
		                fusedev_get_data(fdi.fdev)->daemoncred, cred)))
		{
			struct fuse_mnt_data *x_fmnt;

			LIST_FOREACH(x_fmnt, &fmnt->slaves_head, slaves_link) {
				if (! (denied = (x_fmnt->mp->mnt_cred->cr_uid != cred->cr_uid)))
					break; 
			}

			if (denied) {
				sx_sunlock(fdi.slock);
				uprintf("Your access is blocked in order to prevent the Fuse daemon spying on you.\n"
				        "To get further, you can do a shared mount, cf. mount_fusefs(8).\n");
				return (EPERM);
			}
		}
	}
	sx_sunlock(fdi.slock);

	if (vp->v_type != vap->va_type) {
		/* stale vnode */
		DEBUG("node #%lu got stale (old type 0x%x, new type 0x%x), kicking...\n",
	        vap->va_fileid, vp->v_type, vap->va_type);
		fuse_vnode_kick(vp, td);
		return (EIO);
	}

	DEBUG("node #%llu, type %d\n", VTOI32(vp), vap->va_type);
#if _DEBUG
	DEBUG2G("\n");	
	vn_printf(vp, " * ");
#endif
	return (0);
}

#define CHECK_RO(mp, vtyp, mode)					\
if ((mode) & VWRITE) {							\
	switch ((vtyp)) {						\
	case VDIR:							\
	case VLNK:							\
	case VREG:							\
		if ((mp)->mnt_flag & MNT_RDONLY) {			\
			DEBUG("no write access (read-only fs)\n");	\
			return (EROFS);					\
		}							\
		break;							\
	default:							\
		break;							\
	}								\
}

static int
fuse_access(struct vop_access_args *ap)
{
	struct vnode *vp = ap->a_vp;
	struct ucred *cred = ap->a_cred;
	struct thread *td = ap->a_td;
	mode_t mode = ap->a_mode;

	int err = 0;	
	struct vattr va;

	/*
	 * Disallow write attempts on read-only file systems; unless the file
	 * is a socket, fifo, or a block or character device resident on the
	 * file system.
	 */

	CHECK_RO(vp->v_mount, vp->v_type, mode);

	if (! (err = VOP_GETATTR(vp, &va, cred, td)))
		err = vaccess(va.va_type, va.va_mode, va.va_uid, va.va_gid,
		              mode, cred, NULL);

	return (err);
}

/*
 * Attribute caching hasn't yet been implemented.
 * However, within one function we don't wanna query attributes
 * several times. Now it's enough pull the attributes once, and throw
 * it into the following routine with various modes.
 */

static __inline int
fuse_access_by_attr(struct mount *mp, struct vattr *vap, mode_t mode, struct ucred *cred)
{
	CHECK_RO(mp, vap->va_type, mode);

	return (vaccess(vap->va_type, vap->va_mode, vap->va_uid, vap->va_gid,
	    mode, cred, NULL));
}

/*
 * Based on sample code of VOP_LOOKUP(9), reiserfs_lookup()
 * of gnu/fs/reiserfs/reiserfs_namei.c  and vfs_cache_lookup()
 */

int
/* fuse_lookup(struct vop_cachedlookup_args *ap) */
fuse_lookup(struct vop_lookup_args *ap)
{
        struct vnode *dvp         = ap->a_dvp;
        struct vnode **vpp        = ap->a_vpp;
        struct componentname *cnp = ap->a_cnp;
	int nameiop = cnp->cn_nameiop;
	struct thread *td = cnp->cn_thread;
	struct ucred *cred = cnp->cn_cred;
	int flags = cnp->cn_flags;
	int lockparent = flags & LOCKPARENT;
	int wantparent = flags & WANTPARENT;
	int islastcn = flags & ISLASTCN;

	int err = 0;
	int lookup_err = 0;
	struct vnode *vp = NULL;
	struct vnode *pdp = NULL;
	struct fuse_dispatcher fdi;
	struct fuse_attr *fattr = NULL;
	struct vattr dva;
	enum fuse_opcode op;
	uint64_t nid;

	/* general stuff, based on vfs_cache_lookup */

#if _DEBUG2G
	DEBUG2G("root node:\n");
	vn_printf(((struct fuse_mnt_data *)dvp->v_mount->mnt_data)->rvp, " * ");
#endif

	if (dvp->v_type != VDIR) {
		DEBUG("vnode #%llu of vtype %d is not a dir\n", VTOI32(dvp),
		      dvp->v_type); 
		return (ENOTDIR);
	}

	if ((flags & ISLASTCN) && (dvp->v_mount->mnt_flag & MNT_RDONLY) &&
	    (cnp->cn_nameiop == DELETE || cnp->cn_nameiop == RENAME))
		return (EROFS);

	/*
	 * We do access check prior to doing anything else only in the case
	 * when we are at fs root (we'd like to say, "we are at the first
	 * component", but that's not exactly the same... nevermind).
	 * See further comments at further access checks.
	 */

	if (
#if NO_EARLY_PERM_CHECK_HACK
	    1
#else
	    dvp->v_vflag & VV_ROOT
#endif
	   ) {
		if ((err = VOP_ACCESS(dvp, VEXEC, cred, td)))
			return (err);
	}

	DEBUG2G("lookup in #%llu, with flags 0x%x\n", VTOI32(dvp), flags);

	/* fetching data from "storage" */

	if (FUSE_NAME_MAX < cnp->cn_namelen)
		return (ENAMETOOLONG);

	if (flags & ISDOTDOT) {
		DEBUG2G("dotdot lookup!\n");
		GETPARENT(pdp, dvp); 
		KASSERT(pdp, ("dotdot lookup attempted without knowing parent")); 
		nid = VTOI(pdp);
		fdi.iosize = 0;
		op = FUSE_GETATTR;
	} else if (cnp->cn_namelen == 1 && *(cnp->cn_nameptr) == '.') {
		DEBUG2G("dot lookup!\n");
		nid = VTOI(dvp);
		fdi.iosize = 0;
		op = FUSE_GETATTR;
	} else {
		nid = VTOI(dvp);
		fdi.iosize = cnp->cn_namelen + 1;
		op = FUSE_LOOKUP;
	}


	if ((err = fdisp_get_vopdata(&fdi, dvp->v_mount)))
		return (err);
	fdisp_prepare_msg(&fdi, op, nid, td, cred);

	if (op == FUSE_LOOKUP) {
		memcpy(fdi.indata, cnp->cn_nameptr, cnp->cn_namelen);
		((char *)fdi.indata)[cnp->cn_namelen] = '\0';
		DEBUG2G("standard lookup, looking up %s\n", (char *)fdi.indata);
	}

	lookup_err = fdisp_wait_answ(&fdi);
	if (lookup_err && (
	    (! fdi.answ_stat) || /* this means messaging error */
            lookup_err != ENOENT /* daemon reported other error than "legal" ENOENT */
            )) {
		/*
		 * There is error but not lookup related actually
		 * (messaging error)
		 */
		DEBUG("lookup failed b/c messaging crap\n");
		return (lookup_err);
	}

	/* analyzing answer */

	/*
	 * Now we got the answer and filtered out the crap, too, so we know that
	 * "found" iff lookup_err == 0
	 */

	if (lookup_err) {
		DEBUG("looked up thingy not found\n");
		if ((nameiop == CREATE || nameiop == RENAME)
		 && islastcn
		 /* && directory dvp has not been removed */) {
			DEBUG("create/rename -- we have to make it\n");
			/*
			 * Check for write access on directory.
			 */
			
			/*
			 * Possibly record the position of a slot in the
			 * directory large enough for the new component name.
			 * This can be recorded in the vnode private data for
			 * dvp. Set the SAVENAME flag to hold onto the
			 * pathname for use later in VOP_CREATE or VOP_RENAME.
			 */
			cnp->cn_flags |= SAVENAME;
			if (!lockparent)
				/*
				 * Note that the extra data recorded above is
				 * only useful if lockparent is specified.
				 */
				VOP_UNLOCK(dvp, flags, td);
			
			err = EJUSTRETURN;
			goto out;
		}

		/*
		 * Consider inserting name into cache.
		 */

		/*
		 * No we can't use negative caching, as the fs
		 * changes are out of our control.
		 * False positives' falseness turns out just as things
		 * go by, but false negatives' falseness doesn't.
		 * (and aiding the caching mechanism with extra control
		 * mechanisms comes quite close to beating the whole purpose
		 * caching...)
		 */
#if 0
		if ((cnp->cn_flags & MAKEENTRY) && nameiop != CREATE) {
			DEBUG("inserting NULL into cache\n");
			cache_enter(dvp, NULL, cnp);
		}
#endif
		err = ENOENT;
		goto out;
	} else {
		if (op == FUSE_GETATTR)
			fattr = &((struct fuse_attr_out *)fdi.answ)->attr;
		else {
			fattr = &((struct fuse_entry_out *)fdi.answ)->attr;
			nid = ((struct fuse_entry_out *)fdi.answ)->nodeid;
		}
		
		DEBUG("we found something...\n");
		/*
		 * If deleting, and at end of pathname, return parameters
		 * which can be used to remove file.  If the wantparent flag
		 * isn't set, we return only the directory, otherwise we go on
		 * and lock the inode, being careful with ".".
		 */
		if (nameiop == DELETE && islastcn) {
			DEBUG("something to delete\n");
			/*
			 * Check for write access on directory.
			 */
			err = VOP_ACCESS(dvp, VWRITE, cred, cnp->cn_thread);
			if (err)
				goto out;
		
			if (nid == VTOI(dvp)) {
				VREF(dvp);
				*vpp = dvp;
				goto out;
			}
			/*
			 * XXX generation, also brought to us by LOOKUP, maybe
			 * should be pushed onto some leaf of the vnode; yet I
			 *  don't make up my mind 'till I don't get at attr
			 *  caching...
			 */   
			if ((err = fuse_vget_i(dvp->v_mount, td, nid,
			    IFTOVT(fattr->mode), &vp)))
				goto out;

	
			if ((err = VOP_GETATTR(dvp, &dva, cred, td)))
				goto out;

	
			if ((dva.va_mode & S_ISTXT)
			 && cred->cr_uid != 0
			 && cred->cr_uid != dva.va_uid
			 && fattr->uid != cred->cr_uid) {
				vput(vp);
				err = EPERM;
				goto out;
			}
			*vpp = vp;
			if (!lockparent)
				VOP_UNLOCK(dvp, flags, td);
		
			goto out;
		}

		/*
		 * If rewriting (RENAME), return the inode and the
		 * information required to rewrite the present directory
		 * Must get inode of directory entry to verify it's a
		 * regular file, or empty directory.
		 */
		if (nameiop == RENAME && wantparent && islastcn) {
			DEBUG("something to rename...\n");
			err = VOP_ACCESS(dvp, VWRITE, cred, cnp->cn_thread);
			if (err)
				goto out;
		
			/*
			 * Check for "."
			 */
			if (nid == VTOI(dvp)) {
				err = EISDIR;
				goto out;
			}
		
			if ((err = fuse_vget_i(dvp->v_mount, td, nid, IFTOVT(fattr->mode), &vp)))
				goto out;
			*vpp = vp;
			/*
			 * Save the name for use in VOP_RENAME later.
			 */
			cnp->cn_flags |= SAVENAME;
			if (!lockparent)
				VOP_UNLOCK(dvp, flags, td);
		
			goto out;
		}

		DEBUG("we peacefully found that file\n");

		if (flags & ISDOTDOT) {
			vn_lock(pdp, LK_EXCLUSIVE | LK_RETRY, td);
			VREF(pdp);
			*vpp = pdp;
		} else if (nid == VTOI(dvp)) {
			VREF(dvp); /* We want ourself, ie "." */
			*vpp = dvp;
		} else {
			if ((err = fuse_vget_i(dvp->v_mount, td, nid, IFTOVT(fattr->mode), &vp)))
				goto out;
			if (vp->v_type == VDIR)
				SETPARENT(vp, dvp);
			*vpp = vp;
		}

		/* Insert name into cache if appropriate. */

		/*
		 * Nooo, caching is evil. With caching, we can't avoid stale
		 * information taking over the playground (cached info is not
		 * just positive/negative, it does have qualitative aspects,
		 * too). And a (VOP/FUSE)_GETATTR is always thrown anyway, when
		 * walking down along cached path components, and that's not
		 * any cheaper than FUSE_LOOKUP. This might change with
		 * implementing kernel side attr caching, but... In Linux,
		 * lookup results are not cached, and the daemon is bombarded
		 * with FUSE_LOOKUPS on and on. This shows that by design, the
		 * daemon is expected to handle frequent lookup queries
		 * efficiently, do its caching in userspace, and so on.
		 *
		 * So just leave the name cache alone.
		 */

		/*
		 * Well, now I know, Linux caches lookups, but with a
		 * timeout... So it's the same thing as attribute caching:
		 * we can deal with it when implement timeouts.
		 */	
#if 0
		if (cnp->cn_flags & MAKEENTRY) {
			DEBUG("...caching\n");
			cache_enter(dvp, *vpp, cnp);
		}
#endif
	}
out:
	if (! lookup_err) {
		DEBUG("as the looked up thing was simply found, the cleanup is left for us\n");

		if (err) {
			DEBUG("though inode found, err exit with no vnode\n");
			if (op == FUSE_LOOKUP)
				fuse_send_forget(dvp->v_mount, td, cred, nid, 1,
				                 &fdi);
			return (err);
		} else {
			if (op == FUSE_LOOKUP) {
				((struct fuse_vnode_data *)(*vpp)->v_data)->nlookup++;
#if _DEBUG
				DEBUG2G("nlookup bumped:\n");
				vn_printf(*vpp, " * ");
#endif
			}
#if ! NO_EARLY_PERM_CHECK_HACK
			if (! islastcn) {
				/* We have the attributes of the next item
				 * *now*, and it's a fact, and we do not have
				 * to do extra work for it (ie, beg the
				 * daemon), and it neither depends on such
				 * accidental things like attr caching. So the
				 * big idea: check credentials *now*, not at
				 * the beginning of the next call to lookup.
				 *
				 * The first item of the lookup chain (fs root)
				 * won't be checked then here, of course, as
				 * its never "the next". But go and see that
				 * the root is taken care about at the very
				 * beginning of this function.
				 *
				 * Now, given we want to do the access check
				 * this way, one might ask: so then why not do
				 * the access check just after fetching the
				 * inode and its attributes from the daemon?
				 * Why bother with producing the corresponding
				 * vnode at all if something is not OK? We know
				 * what's the deal as soon as we get those
				 * attrs... There is one bit of info though not
				 * given us by the daemon: whether his response
				 * is authorative or not... His response should
				 * be ignored if something is mounted over the
				 * dir in question. But that can be known only
				 * by having the vnode...
				 */
				struct vattr va;
				DEBUG("the early perm check hack\n");
				fat2vat(dvp->v_mount, fattr, &va);
				if (va.va_type != VDIR && va.va_type != VLNK)
					err = ENOTDIR;
				if (! err && ! (*vpp)->v_mountedhere)
					err = fuse_access_by_attr(dvp->v_mount, &va, VEXEC, cred);
				if (err) {
					if (va.va_type == VLNK)
						DEBUG("weird, permission error with a symlink?\n");
					vput(*vpp);
					*vpp = NULL;
				}
			}
#endif
		}
			
		ticket_drop(fdi.tick);
		sx_sunlock(fdi.slock);
	}

	DEBUG("returning with %d\n", err);
#if _DEBUG2G
	if (!err) {
		DEBUG2G("looked up node:\n");
		vn_printf(*vpp, " * ");
	}
#endif
#if _DEBUG2G
	DEBUG2G("root node:\n");
	vn_printf(((struct fuse_mnt_data *)dvp->v_mount->mnt_data)->rvp, " * ");
#endif
	return (err);
}

static int
iterate_filehandles(struct vnode *vp, struct thread *td, struct ucred *cred,
		    int mode, fuse_metrics_t fmetr, void *param)
{
	struct fuse_filehandle *fufh, *fufhxxx;
	struct fuse_vnode_data *fvdat = vp->v_data;
	int rv = 0;

	RECTIFY_TDCR(td, cred);

	LIST_FOREACH_SAFE(fufh, &fvdat->fh_head, fh_link, fufhxxx) {
		if ((rv = fmetr(vp, td, cred, mode, fufh, param)))
			return (rv); 
	}

	return (0);
}

static int
fuse_standard_metrics(struct vnode *vp, struct thread *td, struct ucred *cred,
                      int mode, struct fuse_filehandle *fufh, void *param) 
{
	DEBUG2G("fufh->mode %x, mode %x\n", fufh->mode, mode);
	if (fufh->cred->cr_uid == cred->cr_uid &&
	    fufh->cred->cr_rgid == cred->cr_rgid &&
	    fufh->pid == td->td_proc->p_pid &&
	    fufh->mode == mode) { 
		*(struct fuse_filehandle **)param = fufh;
		fufh->useco++;
		return (-1);
	}

	return (0);
}
	
static struct fuse_filehandle *
get_filehandle(struct vnode *vp, struct thread *td, struct ucred *cred,
               int mode)
{
	struct fuse_vnode_data *fvdat = vp->v_data;
	struct fuse_dispatcher fdi;
	struct fuse_open_in *foi;
	struct fuse_open_out *foo;
	int err = 0;
	struct fuse_filehandle *fufh = NULL;

	mode &= mode & ~(O_CREAT | O_EXCL | O_NOCTTY | O_TRUNC);

	RECTIFY_TDCR(td, cred);

	sx_slock(&fvdat->fh_lock);
	err = iterate_filehandles(vp, td, cred, mode, fuse_standard_metrics,
	                          &fufh);
	sx_sunlock(&fvdat->fh_lock);

	if (err == -1) {
		DEBUG2G("suitable fh of vnode #%llu found\n", VTOI32(vp));
		return (fufh);
	}
	else if (err)
		return (NULL);

	DEBUG2G("we need to fetch a new filehandle for vnode #%llu\n", VTOI32(vp));

	fdi.iosize = sizeof(*foi);
	if ((err = fdisp_prepare_all(&fdi,
             (vp->v_type == VDIR) ? FUSE_OPENDIR : FUSE_OPEN, vp, td, cred)))
		return (NULL);

	foi = fdi.indata;
	foi->flags = OFLAGS(mode);

	if ((err = fdisp_wait_answ(&fdi))) {
		DEBUG2G("err %d\n", err);
		return (NULL);
	}
	foo = fdi.answ;

	MALLOC(fufh, struct fuse_filehandle *, sizeof(*fufh), M_FUSEFH,
	       M_WAITOK | M_ZERO);

	fufh->fh_id = foo->fh;

	ticket_drop(fdi.tick);
	sx_sunlock(fdi.slock);

	fufh->mode = mode;
	fufh->cred = crhold(cred);
	fufh->pid = td->td_proc->p_pid;
	fufh->useco = 1;
	fufh->fp = NULL;

	sx_xlock(&fvdat->fh_lock);
	LIST_INSERT_HEAD(&fvdat->fh_head, fufh, fh_link);
	sx_xunlock(&fvdat->fh_lock);
	fvdat->fh_counter++;

	return (fufh);
}

static int
fuse_open(struct vop_open_args *ap)
{
	struct thread *td = ap->a_td;
	struct vnode *vp = ap->a_vp;
	int fdidx = ap->a_fdidx;
	struct ucred *cred = ap->a_cred;
	int mode = ap->a_mode;

	struct fuse_filehandle *fufh = NULL;
	struct fuse_vnode_data *fvdat = vp->v_data;
#if DIRECTIO_FOR_DIRS
	struct ucred *clonecrp;
#endif
	struct fuse_dispatcher fdi;
	struct fuse_open_in *foi;
	struct fuse_open_out *foo;
	int err = 0;
	struct file *fp = NULL;
	int keep_cache;

	if (! vp->v_object)
		/* The "if" here is just to avoid needless getattr'ing */
		vnode_create_vobject(vp, 0, td);

#if DIRECTIO_FOR_DIRS
	if (vp->v_type == VDIR) {
		if (fdidx < 0)
			return (EINVAL);
	}	
#endif

	if (vp->v_type == VDIR)
		mode = FREAD;
	else
		mode &= ~(O_CREAT | O_EXCL | O_NOCTTY | O_TRUNC);

	/*
	 * Contrary to that how it used to be, now we go on even if it's an
	 * internal (fileless) open: to match the "specs", we must get the
	 * keep_cache information (and act according to it).
	 */
	if (fdidx >= 0) {
		/*
		 * That certain "pretty disgustingly long chain"
		 * by which we can put our hands of the file struct
		 * assinged to this call.
		 */
		fp = td->td_proc->p_fd->fd_ofiles[fdidx];
		if (! fp)
			panic("nonneg file desc passed to us but no file there");

		DEBUG2G("fp->f_flag 0x%x, open mode %d\n", fp->f_flag, mode);
	}
	fdi.iosize = sizeof(*foi);
	if ((err = fdisp_prepare_all(&fdi,
	     (vp->v_type == VDIR) ? FUSE_OPENDIR : FUSE_OPEN, vp, td, cred)))
		return (err);

	foi = fdi.indata;
	foi->flags = OFLAGS(mode);

	if ((err = fdisp_wait_answ(&fdi)))
		return (err);
	foo = fdi.answ;

	MALLOC(fufh, struct fuse_filehandle *, sizeof(* fufh), M_FUSEFH,
	       M_WAITOK | M_ZERO);

	fufh->fh_id = foo->fh;

	if (fp)
		fp->f_flag = mode;
	if (foo->open_flags & FOPEN_DIRECT_IO) {
		if (fp)
			/* This seems not to be permanent */
			fp->f_flag |= O_DIRECT;
		fufh->flags |=  FUSEFH_DIRECTIO;
	}

	keep_cache = foo->open_flags & FOPEN_KEEP_CACHE;

	ticket_drop(fdi.tick);
	sx_sunlock(fdi.slock);

#if DIRECTIO_FOR_DIRS
	if (vp->v_type == VDIR) {
		DEBUG("coloring file #%d\n", fdidx);
		/*
		 * ~~~ File struct coloring for VOP_READDIR ~~~
		 *
		 * [We used to think that] we *must* know the file struct via
		 * which a VOP_READDIR happens (we will need to pass the
		 * filehandle -- given us now, at open time -- to FUSE_READDIR).
		 * While it's not a problem with VOP_READ, where the file's own
		 * f_op-s are used (though by default they just call the vnode's
		 * corresponding vn_op, they can be set to anything),
		 * in case of VOP_READDIR there is no corresponding f_op, and
		 * the vnode's version is invoked directly (see getdirentries()
		 * in kern/vfs_syscalls.c).
		 * What the vnode will see from the file is only its f_cred
		 * field, by which the file is hardly identifiable, as it's
		 * just a reference to the thread's cred structure (canonically
		 * pointed to by the td_ucred field).  So here at open time we
		 * "colorize" the files such that fuse_readdir will be able to
		 * distinguish them.
		 * That is, being not afraid of a healthy dose of
		 * hackityhackity hack, we replace f_cred with a clone devoted
		 * for this purpose. Then when fuse_readdir comes on to the
		 * scene, he can use the passed on cred struct like the prince
		 * uses Cinderella's slippers: he walks over the list of open
		 * files, and tries the cred against each of them, and he will
		 * pick the one which fits.
		 * 
		 * XXX: does this break anything? eg. imagine that the thread's
		 * creds -- then the clone used by the file won't be updated...
		 * Forunately, it seems that the file's creds only used for
		 * access check only against the vnode it belongs to, and here
		 * with Fuse we don't use file creds in the vnops at all
		 * (except for this hack). MAC stuff should be looked at a bit
		 * more closer, though... there are a buch of 
		 * "mac_check_foo(vp, fp)" type ops, and though I'll be
		 * surprised if that fp weren't one of that vp, I can't
		 * exclude this now. (Anyway, I don't implement MAC related
		 * things now, so my guess that MAC-related syscalls against
		 * us will fall with EOPNOTSUPP nevertheless...
		 */
		clonecrp = crdup(fp->f_cred);
		crfree(fp->f_cred);
		fp->f_cred = clonecrp;
	}
#endif
	
	/*
	 * Types: VNON, VREG, VDIR, VBLK, VCHR, VLNK, VSOCK, VFIFO, VBAD 
	 * XXX: should we let the fs take over spec files or let the OS
	 *  directly handle their IO?
	 */
	/* This is neither got right, nor settled */
	#if 0 
		if (! (vp->v_type == VFIFO || vp->v_type == VSOCK))
			fp->f_ops = &fuse_fileops;
	#endif

	fufh->mode = mode;
	fufh->cred = crhold(cred);
	fufh->pid = td->td_proc->p_pid;
	fufh->useco = 1;
	if (fp) {
		fp->f_ops = &fuse_fileops;
		fufh->fp = fp;
		fp->f_data = fufh;
	}

#if _DEBUG
	DEBUG2G("before gc...\n");
	vn_printf(vp, " ");
#endif

	sx_xlock(&fvdat->fh_lock);
	LIST_INSERT_HEAD(&fvdat->fh_head, fufh, fh_link);
	/*
	 * The neat idea is that a cache cleanup is the best time for a gc --
	 * gc'ing unbound filehandles at inactive/reclaim time is not enough,
	 * they can just accumulate while something happens with the file,
	 * given the cache is flushed frequently (otherwise there is no reason
	 * for the system to spawn them, as the data can be read from core).
	 * So then, let's just tie their gc to cache flush...
	 * Luckily, this won't break the POSIX semantics that an unlinked file
	 * is get kept until there is reference to it, because that's done
	 * via the lookup/forget mechanism, and not via filehandles.
	 */
	/*
	 * Now we use unbound filehandles pretty rarely (as read/write of
	 * regular files doesn't rely on them anymore), only during exec
	 * and reading directories. It's pretty much unlikely that one
	 * can be reused in any way, so we drop them unconditionally.
	 */
	DEBUG2G("gc'ing...\n");
	iterate_filehandles(vp, td, cred, 0, release_filehandle, NULL);
	sx_xunlock(&fvdat->fh_lock);
	fvdat->fh_counter++;
#if _DEBUG2G
	DEBUG2G("after gc...\n");
	vn_printf(vp, " ");
#endif

	if (! fp)
		fufh->useco--;

	if (! keep_cache) {
		DEBUG2G("flushing cache\n");
		/*
		 * This is a bit more heavyweight than the analogous function
		 * used in Linux at this point (in their sys/mm/truncate.c,
		 * the description says:
		 * "invalidate_mapping_pages() will not block on IO activity.
		 * It will not invalidate pages which are dirty, locked, under
		 * writeback or mapped into pagetables.") 
		 */
		err = vinvalbuf(vp, 0, td, PCATCH, 0);
	}

	return (err);
}

static int
fuse_send_release(struct vnode *vp, struct thread *td, struct ucred *cred,
                  struct fuse_filehandle *fufh, int flags)
{
	struct fuse_dispatcher fdi;
	struct fuse_release_in *fri;
	int err = 0;
	struct fuse_vnode_data *fvdat = vp->v_data;

	KASSERT(! fufh->fp && fufh->useco == 0, ("active-looking fuse filehandle was attempted to release"));

	fvdat->fh_counter--;
	DEBUG2G("filehandle of vnode #%llu being released, fh counter now is %d\n",
	         VTOI32(vp), fvdat->fh_counter); 
	fdi.iosize = sizeof(*fri);

	if ((err = fdisp_prepare_all(&fdi,
             (vp->v_type == VDIR) ? FUSE_RELEASEDIR : FUSE_RELEASE,
             vp, td, cred)))
		return (err);

	fri = fdi.indata;
	fri->fh = fufh->fh_id;
	fri->flags = OFLAGS(flags);

	fuse_insert_callback(fdi.tick, NULL);
	fuse_insert_message(fdi.tick);
	sx_sunlock(fdi.slock);

	crfree(fufh->cred);
	FREE(fufh, M_FUSEFH);

	return (0);
}

/*
 * XXX for blocking vnode usage, it seems that changing vnops to dead ones
 * is not enough: file ops don't look at the vnode...
 * so as an ad hoc solution we'll insert this macro to the top of file
 * specific ops.
 * Tell me if there is an "official" way to handle this.
 */
#define BREAK_IF_BAD(fp)						\
if (! (fp)->f_vnode->v_data) {						\
	DEBUG("bad fileop on vnode no. %llu\n", VTOI32((fp)->f_vnode));	\
	return (EBADF);							\
}

/*
 * In kern_open() of vfs_syscalls.c, if the f_data slot is not set
 * up at open time, the vnode of the file will be put there.
 * So "f_data == f_vnode" means someting like "the file is not
 * initialized". I don't know how stable this convention is, so relying
 * on it might be somewhat bogus; yet this saves us from on more locking
 * for nothing (when allocating a dummy data field).
 * As a compromise, the respective test is wrapped into a macro, to provide
 * some abstraction...
 */
#define FILE_IS_FAT(fp) ((fp)->f_data && (fp)->f_data != (fp)->f_vnode)


static int
fuse_close_f(struct file *fp, struct thread *td)
{
	struct fuse_filehandle *fufh;
	struct fuse_vnode_data *fvdat;
	int err = 0;

	BREAK_IF_BAD(fp);

	if (! FILE_IS_FAT(fp))
		panic("non-fat file passed to close routine");

	fvdat = fp->f_vnode->v_data;
	fufh = fp->f_data;
	KASSERT(fufh->fp == fp, ("file's filehandle is stolen"));
	DEBUG2G("vnode #%llu, fufh owner %p, useco %d\n",
	        VTOI32(fp->f_vnode), fp, fufh->useco);

	fufh->useco--;
	sx_xlock(&fvdat->fh_lock);
	fufh->fp = NULL;
	if (fufh->useco == 0)
		LIST_REMOVE(fufh, fh_link);
	sx_xunlock(&fvdat->fh_lock);
	fp->f_data = NULL;

	vn_lock(fp->f_vnode, LK_EXCLUSIVE | LK_RETRY, td);	

	if (fufh->useco == 0)
		err = fuse_send_release(fp->f_vnode, td, NULL, fufh,
		                        fp->f_flag & ~O_EXCL);

        if (fp->f_flag & FWRITE)
                fp->f_vnode->v_writecount--;
	vput(fp->f_vnode);
	return (err);
}

int
fuse_read_biobackend(struct vnode *vp, struct fuse_filehandle *fufh,
                     struct uio *uio, struct ucred *cred, struct thread *td,
                     enum fuse_opcode op, fuse_buffeater_t *buffe, void *param)
{
	int biosize;
	struct buf *bp;
	daddr_t lbn;
	int bcount;
	int bbcount;
	int err = 0, n = 0, on = 0;
	struct vattr va; 

	VOP_GETATTR(vp, &va, cred, td); 

	if (uio->uio_resid == 0)
		return (0);

	biosize = vp->v_mount->mnt_stat.f_iosize;
	DEBUG2G("entering loop\n");
	do {
		lbn = uio->uio_offset / biosize;
		on = uio->uio_offset & (biosize - 1);

		DEBUG2G("biosize %d, lbn %d, on %d\n", biosize, (int)lbn, on);

		/*
		 * Obtain the buffer cache block.  Figure out the buffer size
		 * when we are at EOF.  If we are modifying the size of the
		 * buffer based on an EOF condition we need to hold
		 * nfs_rslock() through obtaining the buffer to prevent
		 * a potential writer-appender from messing with n_size.
		 * Otherwise we may accidently truncate the buffer and
		 * lose dirty data.
		 *
		 * Note that bcount is *not* DEV_BSIZE aligned.
		 */

#define BSD_FUSE_PAGENO_BOUND MIN(FUSE_MAX_PAGES_PER_REQ, MAXBSIZE / PAGE_SIZE)

		bcount = biosize * BSD_FUSE_PAGENO_BOUND; 
#if BIOREAD_CONSIDER_FILE_SIZE 
		if (vp->v_type != VDIR) {
			/*
			 * for directories we can't predict the size, in fact
			 * the daemon decides how she will push the entries
			 */
			bcount = MIN(biosize * BSD_FUSE_PAGENO_BOUND, va.va_size - (off_t)lbn * biosize);
			if (bcount < 0)
				bcount = 0;
		}
#endif

		DEBUG("getting block from OS, bcount %d\n", bcount);
		bp = getblk(vp, lbn, bcount, PCATCH, 0, 0);

		if (!bp)
			return (EINTR);

		/*
		 * If B_CACHE is not set, we must issue the read.  If this
		 * fails, we return an error.
		 */

		if ((bp->b_flags & B_CACHE) == 0) {
			bp->b_iocmd = BIO_READ;
			vfs_busy_pages(bp, 0);
			err = fuse_strategy_i(vp, bp, fufh);
#if _DEBUG
			prettyprint(bp->b_data, 48);
			printf("\n");
#if ZERO_PAD_INCOMPLETE_BUFS
			prettyprint(bp->b_data + PAGE_SIZE, 48);
			printf("\n");
#endif
#endif
			if (err) {
				brelse(bp);
				return (err);
			}
		}

		/*
		 * on is the offset into the current bp.  Figure out how many
		 * bytes we can copy out of the bp.  Note that bcount is
		 * NOT DEV_BSIZE aligned.
		 *
		 * Then figure out how many bytes we can copy into the uio.
		 */

		n = 0;
		/*
		 * If we zero pad the buf, bp->b_resid will be 0, so then
		 * just ignore it
		 */
		bbcount = bcount - bp->b_resid;
		if (on < bbcount)
			n = bbcount - on;
		if (n > 0) {
			DEBUG2G("feeding buffeater with %d bytes of buffer %p, saying %d was asked for\n",
			        n, bp->b_data + on, n + (int)bp->b_resid);
#if 0 && _DEBUG
		        prettyprint(bp->b_data + on, n);
			printf("\n");
#endif
		        err = buffe(uio, n + bp->b_resid, bp->b_data + on, n,
			            param);
		}
		brelse(bp);
		DEBUG2G("end of turn, err %d, uio->uio_resid %d, n %d\n",
		      err, uio->uio_resid, n);
	} while (err == 0 && uio->uio_resid > 0 && n > 0);

	return ((err == -1) ? 0 : err);
}

static int
fuse_read_directbackend(struct vnode *vp, struct fuse_filehandle *fufh,
                        struct uio *uio, struct ucred *cred, struct thread *td,
                        enum fuse_opcode op, fuse_buffeater_t *buffe,
                        void *param)
{
	struct fuse_dispatcher fdi;
	struct fuse_read_in *fri;
	int err = 0;

	if (uio->uio_resid == 0)
		return (0);

	DEBUG("bug daemon for food\n");

	if ((err = fdisp_get_vopdata(&fdi, vp->v_mount)))
		return (err);
	
	while (uio->uio_resid > 0) {
		fdi.iosize = sizeof(*fri);
		fdisp_prepare_msg(&fdi, op, VTOI(vp), td, cred);
	
		fri = fdi.indata;
		fri->fh = fufh->fh_id;
		fri->offset = uio->uio_offset;
		fri->size = MIN(uio->uio_resid,
		                PAGE_SIZE * FUSE_MAX_PAGES_PER_REQ);
	
		DEBUG2G("fri->fh %llu, fri->offset %d, fri->size %d\n",
		        (unsigned long long)fri->fh, (int)fri->offset, fri->size);
		if ((err = fdisp_wait_answ(&fdi)))
			goto out;
	
		DEBUG2G("%d bytes asked for from offset %d, passing on the %d we got\n",
		        uio->uio_resid, (int)uio->uio_offset, (int)fdi.iosize);

		if ((err = buffe(uio, fri->size, fdi.answ, fdi.iosize, param)))
			break;

		ticket_refresh(fdi.tick);
	}

	ticket_drop(fdi.tick);
	sx_sunlock(fdi.slock);

out:
	return ((err == -1) ? 0 : err);
}

static int
fuse_read_f(struct file *fp, struct uio *uio, struct ucred *cred, int flags,
            struct thread *td)
{
	int err = 0;
	struct fuse_filehandle *fufh = fp->f_data;

	BREAK_IF_BAD(fp);

	if (! FILE_IS_FAT(fp))
		panic("non-fat file passed to read routine");

	vn_lock(fp->f_vnode, LK_EXCLUSIVE | LK_RETRY, td);

	if (fp->f_vnode->v_type == VBAD) {
#if _DEBUG2G
		DEBUG2G("bad file\n");
		vn_printf(fp->f_vnode, " * ");
#endif
		VOP_UNLOCK(fp->f_vnode, 0, td);
		return (EBADF);
	}

	if ((flags & FOF_OFFSET) == 0)
		uio->uio_offset = fp->f_offset;

	DEBUG2G("fp->f_flag 0x%x\n", fp->f_flag);

	/*
	 * Ideally, when the daemon asks for direct io at open time, the
	 * standard file flag should be set according to this, so that would
	 * just change the default mode, which later on could be changed via
	 * fcntl(2).
	 * But this doesn't work, the O_DIRECT flag gets cleared at some point
	 * (don't know where). So to make any use of the Fuse direct_io option,
	 * we hardwire it into the file's private data (similarly to Linux,
	 * btw.).
	 */
	if (fp->f_flag & O_DIRECT ||
	    fufh->flags & FUSEFH_DIRECTIO) {
		DEBUG2G("direct read of vnode %llu via file handle %llu\n",
		         VTOI32(fp->f_vnode), (unsigned long long)fufh->fh_id);
		err = fuse_read_directbackend(fp->f_vnode, fufh, uio, cred, td,
		                              FUSE_READ, fuse_std_buffeater,
		                              NULL);
	} else {
		DEBUG2G("buffered read of vnode %llu\n", VTOI32(fp->f_vnode));
		err = fuse_read_biobackend(fp->f_vnode, fufh, uio, cred, td,
		                           FUSE_READ, fuse_std_buffeater, NULL);
	}

	if ((flags & FOF_OFFSET) == 0)
		fp->f_offset = uio->uio_offset;
	fp->f_nextoff = uio->uio_offset;
	VOP_UNLOCK(fp->f_vnode, 0, td);

	return (err);
}
	
static int
fuse_std_buffeater(struct uio *uio, size_t reqsize, void *buf, size_t bufsize, void *param)
{
		int err;

		if ((err = uiomove(buf, MIN(reqsize, bufsize), uio)))
			return (err);

		if (bufsize < reqsize)
			return (-1);

		return (0);
}

#define DIRCOOKEDSIZE FUSE_DIRENT_ALIGN(FUSE_NAME_OFFSET + MAXNAMLEN + 1)

static int
fuse_readdir(struct vop_readdir_args *ap)
{
	struct vnode *vp = ap->a_vp;
	struct ucred *cred = ap->a_cred;
	struct uio *uio = ap->a_uio;

	struct thread *td = curthread;
	int err = 0;
	struct fuse_iov cookediov;
	struct fuse_filehandle *fufh;
#if DIRECTIO_FOR_DIRS
	struct filedesc *fdp = td->td_proc->p_fd;
	struct file **fpp;
	struct file *fp = NULL;
	int found = -1; 
	int f2 = -1;
	int i;

	/* see file coloring in fuse_open() to understand this */

	DEBUG("searching for colored file\n");

	FILEDESC_LOCK_FAST(fdp);
	fpp = fdp->fd_ofiles;
	for (i=0; i < fdp->fd_nfiles; i++) {
		/*
		 * As I don't see why would the list of files
		 * be continuous (one can be
		 * closed while the other is still
		 * partying...), we keep on going even after a null entry,
		 * unlike similar code in fs/fdescfs/fdesc_vnops.c's  
		 * fdesc_readdir().
		 */
		if (*fpp && (*fpp)->f_cred == cred) {
			if (found == -1) {
				fp = *fpp;
				found = i;
			} else {
				f2 = i;
				break;
			}
		}
		fpp++;
	} 
	FILEDESC_UNLOCK_FAST(fdp);

	if (found == -1) {
		KASSERT(0, ("colored file struct has been lost"));
		return (EBADF);
	}
	if (f2 != -1) {
		KASSERT(0, ("file coloring has failed: fd %d and %d look the same to me", found, f2));
		return (EBADF);
	}

	DEBUG("colored file found with fd %d\n", found);

	KASSERT(FILE_IS_FAT(fp), ("DIRECTIO_FOR_DIRS is used but got dummy file"));
	fufh = fp->f_data;
#endif

#if ! DIRECTIO_FOR_DIRS
	if ( ! (fufh = get_filehandle(vp, td, cred, FREAD))) {
		DEBUG2G("fetching filehandle failed\n");
		return (EIO);
	}
#endif
	/*
	 * In "most cases" this will do and won't have to resize the buffer
	 * (that is, if the daemon has a "tight" view of the dir entries,
	 * which is the case if userspace uses the old API or the new one
	 * in the analogous way; by our terminology, this means
	 * freclen == bytesavail)
	 * (See fuse_dir_buffeater).
	 */
	fuse_iov_init(&cookediov, DIRCOOKEDSIZE);
	/*
	 * We tried hard to use bio, but offsety readdir can't be handled
	 * properly that way -- the offset field of fuse_dirents can't be
	 * mapped to an offset of a bio buffer
	 */
	err = fuse_read_directbackend(vp, fufh, uio, cred, td, FUSE_READDIR,
	                              fuse_dir_buffeater, &cookediov);
	fuse_iov_teardown(&cookediov);
#if ! DIRECTIO_FOR_DIRS
	fufh->useco--;
#endif

	return (err);
}

/* A hack for being able to please the GENERIC_DIRSIZ macro */

struct pseudo_dirent {
	uint32_t d_namlen;
};

static int
fuse_dir_buffeater(struct uio *uio, size_t reqsize, void *buf, size_t bufsize, void *param)
{
	/*
	 * The daemon presents us with a virtual geometry when reading dirents.
	 * This info is stored in the "off" field of fuse_dirent (which is the
	 * struct we can read from her). "off" shows the absolut position of
	 * the next entry (by definition). So I might pull 40 actual bytes from
	 * the daemon, that might count as 60 by her virtual geometry, and when
	 * I translate fuse_dirents to POSIX dirents, I have 20 bytes to pass
	 * to the userspace reader.
	 *	
	 * How to keep these in sync? We don't want to make the general read
	 * routine complex (to which here we serve a background, and which
	 * naively treats reading by a liner logic). So we artifically inflate
	 * dirents: we pad them with as much zeros as we need them to get to
	 * next offset (as the deamon declared), and we pass it to the reader
	 * thusly (this will work as fuse dirents are bigger than std ones). So
	 * we do more IO with the reader than absolutely necessary, but it's
	 * hardly a problem as the expensive thing is reading from the daemon,
	 * not sending data to the reader.
	 *
	 * There is no use to change this simple geometry translation scheme
	 * 'till we do page caching. [We do now, but it's still fine as is.]
	 */
	struct dirent *de;
	struct fuse_dirent *fudge;
	int err = 0;
	size_t freclen;
	int cou = 0;
	int bytesavail;
	struct fuse_iov *cookediov;
	
	KASSERT(bufsize <= reqsize, ("read more than asked for?"));

	DEBUG2G("starting\n");

	/*
	 * Sanity check: if this fails, we would overrun the allocated space
	 * upon entering the loop below, so we'd better leave right now.
	 * If so, we return -1 to terminate reading.
	 */
	if (bufsize < FUSE_NAME_OFFSET)
		return (-1);

	cookediov = param;

	DEBUG2G("entering loop with bufsize %d\n", (int)bufsize);

	/*
	 * Can we avoid infite loops? An infinite loop could occur only if we
	 * leave this function with 0 return value, because otherwise we wont't
	 * be called again. But both 0 exit points imply that some data has
	 * been consumed... because
	 *   1) if a turn is not aborted, it consumes positive amount of data
	 *   2) the 0 jump-out from within the loop can't occur in the first
	 *      turn
	 *   3) if we exit 0 after the loop is over, then at least one turn
	 *      was completed, otherwise we hed exited above with -1.
	 */  

	for (;;) {
#if ZERO_PAD_INCOMPLETE_BUFS
		int i;
		char *c;
#endif

		if (bufsize < FUSE_NAME_OFFSET) {
			err = -1;
			break;
		}

		cou++;
		fudge = (struct fuse_dirent *)buf;
		freclen = FUSE_DIRENT_SIZE(fudge);

		DEBUG("bufsize %d, freclen %d\n", (int)bufsize, (int)freclen);

		/*
		 * Here is an exit condition: we terminate the whole reading
		 * process if a fresh chunk of buffer is already too short to
		 * cut out an entry.
		 * (It it was not the first turn in the loop, nevermind,
		 * return with asking for more)
		 */
		if (bufsize < freclen) {
			err = ((cou == 1) ? -1 : 0);
			break;
		}

#if ZERO_PAD_INCOMPLETE_BUFS
		c = buf;
		for (i = 0; i < FUSE_NAME_OFFSET; i++) {
			if (*c)
				break;
			c++;
		}
		if (!(*c)) {
			err = -1;
			break;
		}
#endif

		/* Sanity checks */

		if (!fudge->namelen || fudge->namelen > MAXNAMLEN) {
			DEBUG2G("bogus namelen %d at turn %d\n",
			        fudge->namelen, cou);
			err = EIO;
			break;
		}

		bytesavail = GENERIC_DIRSIZ((struct pseudo_dirent *)&fudge->namelen); 

		/*
		 * Exit condition 2: if the pretended amount of input is more
		 * than that the userspace wants, then it's time to stop
		 * reading.
		 */  
		if (bytesavail > uio->uio_resid) {
			DEBUG2G("leaving at %d-th item as we have %d bytes but only %d is asked for\n",
			        cou, bytesavail, uio->uio_resid);
			err = -1;
			break;
		}

		fuse_iov_audit(cookediov);
		fuse_iov_adjust(cookediov, bytesavail);
		de = (struct dirent *) cookediov->base;
		de->d_fileno = fudge->ino; /* XXX cast from 64 to 32 bytes */
		de->d_reclen = bytesavail;
		de->d_type = fudge->type; 
		de->d_namlen = fudge->namelen;
		memcpy((char *)cookediov->base + sizeof(struct dirent) - MAXNAMLEN - 1,
		       (char *)buf + FUSE_NAME_OFFSET, fudge->namelen);
		((char *)cookediov->base)[bytesavail] = '\0';

		DEBUG("bytesavail %d, fudge->off %llu, fudge->namelen %d, uio->uio_offset %d, name %s\n",
		      bytesavail, (unsigned long long)fudge->off, fudge->namelen, (int)uio->uio_offset,
		      (char *)cookediov->base + sizeof(struct dirent) - MAXNAMLEN - 1);

		err = uiomove(cookediov->base, cookediov->len, uio);

		if (err)
			break;

		buf = (char*)buf + freclen;
		bufsize -= freclen;
		uio->uio_offset = fudge->off;
	}

	return (err);
}

static int
fuse_readlink(struct vop_readlink_args *ap)
{
	struct vnode *vp = ap->a_vp;
	struct uio *uio = ap->a_uio;

	struct fuse_dispatcher fdi;
	int err;

	if ((err = fdisp_simple_putget(&fdi, FUSE_READLINK, vp, curthread,
	    ap->a_cred)))
		return (err);

	DEBUG("answ %s, push 0x%x, mnton %s\n",
	        (char *)fdi.answ,
	        ((struct fuse_mnt_data *)vp->v_mount->mnt_data)->mntopts & FUSEFS_PUSH_SYMLINKS_IN,
	        vp->v_mount->mnt_stat.f_mntonname);

	if (((char *)fdi.answ)[0] == '/' &&
	    ((struct fuse_mnt_data *)vp->v_mount->mnt_data)->mntopts & FUSEFS_PUSH_SYMLINKS_IN)
		err = uiomove(vp->v_mount->mnt_stat.f_mntonname,
		              strlen(vp->v_mount->mnt_stat.f_mntonname), uio);

	if (! err)
		err = uiomove(fdi.answ, fdi.iosize, uio);

	ticket_drop(fdi.tick);
	sx_sunlock(fdi.slock);

	return (err);
}

static __inline int
fuse_make_entry_req(struct vnode *dvp, struct componentname *cnp,
                    enum fuse_opcode op, void *buf, size_t bufsize,
                    struct fuse_dispatcher *fdip)
{
	int err = 0;

	fdip->iosize = bufsize + cnp->cn_namelen + 1;

	if ((err = fdisp_prepare_all(fdip, op, dvp, curthread, NULL)))
		return (err);

	memcpy(fdip->indata, buf, bufsize);

	memcpy((char *)fdip->indata + bufsize, cnp->cn_nameptr, cnp->cn_namelen);
	((char *)fdip->indata)[bufsize + cnp->cn_namelen] = '\0';

	return (0);
}

static __inline int
fuse_newentry_core(struct mount *mp, struct vnode **vpp, enum vtype vtyp,
                   struct fuse_dispatcher *fdip)
{
	struct fuse_entry_out *feo;
	int err = 0;

	if ((err = fdisp_wait_answ(fdip)))
		return (err);
	
	feo = fdip->answ;

	if (vtyp == IFTOVT(feo->attr.mode))
		/*
		 * We don't wanna enforce that the nodeid we get is really new.
		 * A naive implementation of such a policy wouldn't avoid race
		 * conditions and it's a tolerable fuzz. If the daemon sends an
		 * already existing nodeid, then create will have hardlinking
		 * like effects, and that's it.
		 * The original code seems to have the same policy regarding
		 * this issue.
		 */
		err = fuse_vget_i(mp, curthread, feo->nodeid, vtyp, vpp);
	else {	
		DEBUG("daemon sent us node with inconstent types\n");
		fuse_send_forget(mp, curthread, NULL, feo->nodeid, 1, fdip);
		return (EBADF);
	}

	ticket_drop(fdip->tick);
	sx_sunlock(fdip->slock);

	return (err);
}

static int
fuse_newentry_backend(struct vnode *dvp, struct vnode **vpp,
                      struct componentname *cnp, enum fuse_opcode op,
                      void *buf, size_t bufsize, enum vtype vtyp)
{
	struct fuse_dispatcher fdi;
	int err = 0;

	if ((err = fuse_make_entry_req(dvp, cnp, op, buf, bufsize, &fdi)))
		return (err);

	return (fuse_newentry_core(dvp->v_mount, vpp, vtyp, &fdi));
}

static int
fuse_mknod(struct vop_mknod_args *ap)
{
	struct vnode *dvp = ap->a_dvp;
	struct vnode **vpp = ap->a_vpp;
	struct componentname *cnp = ap->a_cnp;
	struct vattr *vap = ap->a_vap;
	struct fuse_mknod_in fmni; 

	fmni.mode = MAKEIMODE(vap->va_type, vap->va_mode);
	fmni.rdev = vap->va_rdev;

	return fuse_newentry_backend(dvp, vpp, cnp, FUSE_MKNOD, &fmni,
	                             sizeof(fmni), vap->va_type);
}

static int
fuse_create(struct vop_create_args *ap)
{
	struct vattr *vap = ap->a_vap;

	vap->va_rdev = 0;
	return (VOP_MKNOD(ap->a_dvp, ap->a_vpp, ap->a_cnp, vap));
}

static int
fuse_mkdir(struct vop_mkdir_args *ap)
{
	struct vnode *dvp = ap->a_dvp;
	struct vnode **vpp = ap->a_vpp;
	struct componentname *cnp = ap->a_cnp;
	struct vattr *vap = ap->a_vap;
	struct fuse_mkdir_in fmdi; 

	fmdi.mode = MAKEIMODE(vap->va_type, vap->va_mode);

	return fuse_newentry_backend(dvp, vpp, cnp, FUSE_MKDIR, &fmdi,
	                             sizeof(fmdi), VDIR);
}

static int
fuse_link(struct vop_link_args *ap)
{
	struct vnode *dvp = ap->a_tdvp;
	struct vnode *vp = ap->a_vp;
	struct componentname *cnp = ap->a_cnp;

	struct fuse_dispatcher fdi;
	struct fuse_link_in fli; 
	struct fuse_entry_out *feo;
	int err = 0;

	if (dvp->v_mount != vp->v_mount)
		return (EXDEV);

	fli.oldnodeid = VTOI(vp);

	if ((err = fuse_make_entry_req(dvp, cnp, FUSE_LINK, &fli, sizeof(fli),
                                       &fdi)))
		return (err);

	if ((err = fdisp_wait_answ(&fdi)))
		return (err);
	
	feo = fdi.answ;

	if (vp->v_type != IFTOVT(feo->attr.mode)) {
		/*
		 * It's not likely to get here:
		 * the syscall we handle here [link(2)] 
		 * is name and not file based.
		 * Hence a type change in the file to link
		 * usually will be caught in the lookup procedure
		 * between entering the syscall and the call to 
		 * to this routine.
		 * We can neither exclude this happening though:
		 * the daemon might lose a race or can play tricks
		 * with us.
		 * XXX: handling of this case is untested.
		 */
		DEBUG("weird, file changed type during linking\n"); 
		fuse_vnode_kick(vp, NULL);
		err = EBADF;
	}

	ticket_drop(fdi.tick);
	sx_sunlock(fdi.slock);

	return (err);
}

static int
fuse_symlink(struct vop_symlink_args *ap)
{
	struct vnode *dvp = ap->a_dvp;
	struct vnode **vpp = ap->a_vpp;
	struct componentname *cnp = ap->a_cnp;
	char *target = ap->a_target;

	struct fuse_dispatcher fdi;
	size_t len;
	int err = 0;

	/*
	 * Unlike the other creator type calls, here we have to create a message
	 * where the name of the new entry comes first, and the data describing
	 * the entry comes second.
	 * Hence we can't rely on our handy fuse_newentry_backend() routine,
	 * but put together the message manually and just call the core part.
	 */

	len = strlen(target) + 1;
	fdi.iosize = len + cnp->cn_namelen + 1;

	if ((err = fdisp_prepare_all(&fdi, FUSE_SYMLINK, dvp, curthread, NULL)))
		return (err);

	memcpy(fdi.indata, cnp->cn_nameptr, cnp->cn_namelen);
	((char *)fdi.indata)[cnp->cn_namelen] = '\0';
	memcpy((char *)fdi.indata + cnp->cn_namelen + 1, target, len);

	return (fuse_newentry_core(dvp->v_mount, vpp, VLNK, &fdi));
}

static int
fuse_remove_backend(struct vnode *dvp, struct componentname *cnp,
                    enum fuse_opcode op)
{
	struct fuse_dispatcher fdi;
	int err = 0;

	fdi.iosize = cnp->cn_namelen + 1;

	if ((err = fdisp_prepare_all(&fdi, op, dvp, curthread, NULL)))
		return (err);

	memcpy(fdi.indata, cnp->cn_nameptr, cnp->cn_namelen);
	((char *)fdi.indata)[cnp->cn_namelen] = '\0';

	if (! (err = fdisp_wait_answ(&fdi))) {
		ticket_drop(fdi.tick);
		sx_sunlock(fdi.slock);
	}

	return (err);
}

static int
fuse_remove(struct vop_remove_args *ap)
{
	return(fuse_remove_backend(ap->a_dvp, ap->a_cnp, FUSE_UNLINK));
}

static int
fuse_rmdir(struct vop_rmdir_args *ap)
{
	return(fuse_remove_backend(ap->a_dvp, ap->a_cnp, FUSE_RMDIR));
}

static __inline int
fuse_rename_core(struct vnode *fdvp, struct componentname *fcnp,
                 struct vnode *tdvp, struct componentname *tcnp)
{
	struct fuse_dispatcher fdi;
	struct fuse_rename_in *fri;
	int err = 0;

	fdi.iosize = sizeof(*fri) + fcnp->cn_namelen + tcnp->cn_namelen + 2;

	if ((err = fdisp_prepare_all(&fdi, FUSE_RENAME, fdvp, curthread, NULL)))
		return err;

	fri = fdi.indata;
	fri->newdir = VTOI(tdvp);
	memcpy((char *)fdi.indata + sizeof(*fri), fcnp->cn_nameptr,
               fcnp->cn_namelen);
	((char *)fdi.indata)[sizeof(*fri) + fcnp->cn_namelen] = '\0';
	memcpy((char *)fdi.indata + sizeof(*fri) + fcnp->cn_namelen + 1,
               tcnp->cn_nameptr, tcnp->cn_namelen);
	((char *)fdi.indata)[sizeof(*fri) + fcnp->cn_namelen + tcnp->cn_namelen + 1] = '\0';
	
	if (! (err = fdisp_wait_answ(&fdi))) {
		ticket_drop(fdi.tick);
		sx_sunlock(fdi.slock);
	}

	return (err);
}

static int
fuse_rename(struct vop_rename_args *ap)
{
	struct vnode *tvp = ap->a_tvp;
	struct vnode *tdvp = ap->a_tdvp;
	struct vnode *fvp = ap->a_fvp;
	struct vnode *fdvp = ap->a_fdvp;
	struct componentname *tcnp = ap->a_tcnp;
	struct componentname *fcnp = ap->a_fcnp;

	int doingdirectory = 0;
	int err = 0;
	struct thread *td;

	td = curthread;
	
	/*
	 * Check for cross-device rename.
	 */
	if (fvp->v_mount != tdvp->v_mount) {
		DEBUG("cross-device move. Why am I expected to deal with it?!\n");
		err = EXDEV;
		goto out;
	}

	/*
	 * Check if just deleting a link name.
	 */
	if (fvp == tvp) {
		if (fvp->v_type == VDIR) {
			err = EINVAL;
			goto out;
		}

		/*
		 * Release destination.
		 */
		vput(tdvp);
		vput(tvp);

		/*
		 * Delete source. Pretty bizarre stuff.
		 */
		vrele(fdvp);
		vrele(fvp);
		fcnp->cn_flags &= ~MODMASK;
		fcnp->cn_flags |= LOCKPARENT | LOCKLEAF;
		fcnp->cn_nameiop = DELETE;
		VREF(fdvp);
		err = relookup(fdvp, &fvp, fcnp);
		if (err == 0)
			vrele(fdvp);
		return VOP_REMOVE(fdvp, fvp, fcnp);
	}

	err = vn_lock(fvp, LK_EXCLUSIVE | LK_RETRY, td);
	if (err)
		goto out;

	if (fvp->v_type == VDIR) {
		/*
		 * Avoid ".", "..", and aliases of "." for obvious reasons.
		 */
		if ((fcnp->cn_namelen == 1 && fcnp->cn_nameptr[0] == '.')
			|| fdvp == fvp
			|| ((fcnp->cn_flags | tcnp->cn_flags) & ISDOTDOT)) {
			VOP_UNLOCK(fvp, 0, td);
			err = EINVAL;
			goto out;
		}
		doingdirectory = 1;
	}
	/* vrele(fdvp);  ???? */

	/*
	 * If ".." must be changed (ie the directory gets a new
	 * parent) then the source directory must not be in the
	 * directory hierarchy above the target, as this would
	 * orphan everything below the source directory. Also
	 * the user must have write permission in the source so
	 * as to be able to change "..".
	 */
	err = VOP_ACCESS(fvp, VWRITE, tcnp->cn_cred, tcnp->cn_thread);
	VOP_UNLOCK(fvp, 0, td);
	if (err)
		goto out;

#if ! DONT_DO_CHECKPATH
	if (doingdirectory && fdvp != tdvp) {
		/*
		 * Check for pathname conflict.
		 *
		 * In linux, dealing with all these subtleties seems to be a
		 * kernel service, so there is no reliance on the daemon
		 * regarding this issue, and it comes for free.
		 *
		 * Nevertheless, we might decide to pass it over to the daemon
		 * -- she can be expected to cope with this.
		 *
		 * Yet here is an attempt to handle the situation... If it
		 * works, then apart from anything else, it lowers line noise...
		 * [works]
		 */

		/* target dir is locked, we can safely read the parent field */
		struct vnode *pdp;

		GETPARENT(pdp, tdvp);
		if (tdvp == fvp /* || pdp == fvp */) {
			DEBUG("checkpath -- easy win\n");
			err = EINVAL;
			goto out;
		}
		if (pdp /* && pdp != tdvp */) {
			struct vnode *pdp2;
			DEBUG("trying at chenkpath\n");
			do {
				DEBUG("checkpath bumped into node %llu\n", VTOI32(pdp));
				if (pdp->v_vflag & VV_ROOT)
					err = -1;
				if (pdp == fvp) {
					DEBUG("huh, we caught a move-into-subdir-of-itself attempt\n");
					err = EINVAL;
				}
				if (err) {
					if (err == -1)
						err = 0;
					DEBUG("leaving checkpath with err %d\n", err);
					vrele(pdp);
					break;
				}
				vn_lock(pdp, LK_EXCLUSIVE | LK_RETRY, td);
				GETPARENT(pdp2, pdp);
				if (pdp2)
					vref(pdp2);
				VOP_UNLOCK(pdp, 0, td);
				vrele(pdp);
				pdp = pdp2;
			} while (pdp);
			DEBUG("left checkpath with err %d\n", err);
			if (err)
				goto out;
		} else if (! pdp)
			DEBUG("target dir's parent pointer is NULL\n");
	}
#endif
	/*
	 * If the target doesn't exist, link the target to the source and
	 * unlink the source. Otherwise, rewrite the target directory to
	 * reference the source and remove the original entry.
	 */
	if (tvp) {
		/*
		 * Target must be empty if a directory and have no links
		 * to it. Also, ensure source and target are compatible
		 * (both directories, or both not directories).
		 */
		if (tvp->v_type == VDIR) {
			/*
			 * "not empty" misery is handled by the fine daemon
			 * (in nulix too)
			 */
			if (!doingdirectory) {
				err = ENOTDIR;
				goto out;
			}
			/*
			 * Update name cache since directory is going away.
			 */
			/* cache_purge(tdvp); */
		} else if (doingdirectory) {
			err = ENOTDIR;
			goto out;
		}

	}

/* Fuse-specific code */

	if ((err = fuse_rename_core(fdvp, fcnp, tdvp, tcnp)))
		goto out;

#ifdef INVARIANTS
	/*
	 * This is here just to stress-test the assertion
	 * of GETPARENT giving no null in fuse_lookup()
	 */
	if (doingdirectory) {
		vn_lock(fvp, LK_EXCLUSIVE | LK_RETRY, td);
		SETPARENT(fvp, NULL);
		VOP_UNLOCK(fvp, 0, td);
	}
#endif
	
out:
	if (tdvp == tvp)
		vrele(tdvp);
	else
		vput(tdvp);
	if (tvp)
		vput(tvp);
	vrele(fdvp);
	vrele(fvp);
	return err;
}

static int
fuse_setattr(struct vop_setattr_args *ap)
{
	struct vattr *vap = ap->a_vap;
	struct vnode *vp = ap->a_vp;
	struct ucred *cred = ap->a_cred;
	struct thread *td = ap->a_td;

	int err = 0;
	struct fuse_dispatcher fdi;
	struct fuse_setattr_in *fsai;
	
	/*
	 * Check for unsettable attributes.
	 */
	if ((vap->va_type != VNON) || (vap->va_nlink != VNOVAL) ||
	    (vap->va_fsid != VNOVAL) || (vap->va_fileid != VNOVAL) ||
	    (vap->va_blocksize != VNOVAL) || (vap->va_rdev != VNOVAL) ||
	    ((int)vap->va_bytes != VNOVAL) || (vap->va_gen != VNOVAL)
#if 0
	     ||
	    /* we don't have the semantics of dealing with these flags */
	    /* it's still used at permission checking for time setting */
	    (vap->va_flags != VNOVAL)
#endif
	    )
		return (EINVAL);

	fdi.iosize = sizeof(*fsai);
	if ((err = fdisp_prepare_all(&fdi, FUSE_SETATTR, vp, td, cred)))
		return (err);
	fsai = fdi.indata;
	/* This is probably unnecessary but makes me feel better */
	bzero(fsai, sizeof(*fsai));

	if (vap->va_uid != (uid_t)VNOVAL) {
		if (vp->v_mount->mnt_flag & MNT_RDONLY)
			return (EROFS);
		fsai->attr.uid = vap->va_uid;
		fsai->valid |= FATTR_UID;
	}

 	if (vap->va_gid != (gid_t)VNOVAL) {
		if (vp->v_mount->mnt_flag & MNT_RDONLY)
			return (EROFS);
		fsai->attr.gid = vap->va_gid;
		fsai->valid |= FATTR_GID;
	}
		
	if (vap->va_size != VNOVAL) {
		/*
		 * Disallow write attempts on read-only file systems;
		 * unless the file is a socket, fifo, or a block or
		 * character device resident on the file system.
		 */
		switch (vp->v_type) {
		case VDIR:
			err = EISDIR;
			goto out;
		case VLNK:
		case VREG:
			if (vp->v_mount->mnt_flag & MNT_RDONLY) {
				err = EROFS;
				goto out;
			}
			break;
		default:
			break;
		}
		fsai->attr.size = vap->va_size;
		fsai->valid |= FATTR_SIZE;	
	}

	if (vap->va_mode != (mode_t)VNOVAL) {
		if (vap->va_mode & S_IFMT)
			DEBUG("fuse_setattr -- weird: format bits in mode field, 0%o\n", vap->va_mode);
		if (vp->v_mount->mnt_flag & MNT_RDONLY)
			return (EROFS);
		fsai->attr.mode = vap->va_mode & ALLPERMS;
		fsai->valid |= FATTR_MODE;
		
	}

	if (vap->va_atime.tv_sec != VNOVAL || vap->va_mtime.tv_sec != VNOVAL) {
		if (vp->v_mount->mnt_flag & MNT_RDONLY)
			return (EROFS);

		/* There is no use for us to check permissions.
		 * The daemon should now what she's doing.
		 */
#if 0
		struct vattr oldva;

		/*
		 * From utimes(2):
		 * If times is NULL, ... The caller must be the owner of
		 * the file, have permission to write the file, or be the
		 * super-user.
		 * If times is non-NULL, ... The caller must be the owner of
		 * the file or be the super-user.
		 */

		if ((err = VOP_GETATTR(vp, &oldva, cred, td)))
			goto out;
		
		if ((err = fuse_access_by_attr(vp->v_mount, &oldva, VADMIN, cred)) &&
		    ((vap->va_vaflags & VA_UTIMES_NULL) == 0 ||
		    (err = fuse_access_by_attr(vp->v_mount, &oldva, VWRITE, cred))))
			goto out;
#endif

		/* XXX: fuse code says, the (atime, mtime) pair should be set
		 * atomically. On the contrary, BSD code doesn't seem to care
		 * about such a constraint. Is this a fuseism or a linuxism?
		 * Should we enforce the constraint?
		 */
		if (vap->va_atime.tv_sec != VNOVAL) {
			/* XXX on some platforms 32 -> bits cast ? */ 
			fsai->attr.atime = vap->va_atime.tv_sec;
			fsai->attr.atimensec = vap->va_atime.tv_nsec;
			fsai->valid |=  FATTR_ATIME;
		}
		if (vap->va_mtime.tv_sec != VNOVAL) {
			fsai->attr.mtime = vap->va_mtime.tv_sec;
			fsai->attr.mtimensec = vap->va_mtime.tv_nsec;
			fsai->valid |=  FATTR_MTIME;
		}
	}

	if ((err = fdisp_wait_answ(&fdi)))
		return (err);

	if (vp->v_type != IFTOVT(((struct fuse_attr_out *)fdi.answ)->attr.mode)) {
		fuse_vnode_kick(vp, td);
		err = EIO;
	}

out:
	ticket_drop(fdi.tick);
	sx_sunlock(fdi.slock);

	return (err);
}

static int 
fuse_fsync_filehandle(struct vnode *vp, struct thread *td, struct ucred *cred,
                      int mode, struct fuse_filehandle *fufh, void *param)
{
	struct fuse_fsync_in *ffsi;
	struct fuse_dispatcher *fdip = param;

	fdip->iosize = sizeof(*ffsi);
	fdip->tick = NULL;
	fdisp_prepare_msg(fdip,
	                  (vp->v_type == VDIR) ? FUSE_FSYNCDIR : FUSE_FSYNC,
	                  VTOI(vp), td, NULL);

	ffsi = fdip->indata;
	ffsi->fh = fufh->fh_id;
	/*
	 * this sets that we wanna sync file data,
	 * not just metadata; a 0 value would have no
	 * counterpart in BSD semantics
	 */
	ffsi->fsync_flags = 1;
  
	fuse_insert_callback(fdip->tick, NULL);
	fuse_insert_message(fdip->tick);

	return (0);
}

static int
fuse_fsync(struct vop_fsync_args *ap)
{
	struct vnode *vp = ap->a_vp;
	struct thread *td = ap->a_td;

	int err = 0;
	struct fuse_dispatcher fdi;
	struct fuse_vnode_data *fvdat;

	DEBUG("got flag %d\n", ap->a_waitfor);

	/*
	 * Usually, the call to vop_stdfsync is expected to do nothing, as we
	 * tend to leave no dirty buffers behind. However, they might show up,
	 * eg., after an interrupted write.
	 * To be able to release the kernel's cleanup-dirty-buffers machinery
	 * this is advised to be inserted here.
	 */

	if ((err = vop_stdfsync(ap)))
		return (err); 

	/*	
	 * In Linux, the fsync syscall handler is file based.
	 * Hence in fuse too -- and this approach does make sense,
	 * for exapmle, when it comes to implementing things like
	 * sshfs: the sshfs userspace fsync handler flushes data
	 * associated with the given filehandle, ie. sftp
	 * messaging thread.
	 * 
	 * On the other hand, BSD fsync is inherently not file
	 * but vnode based: even fsync(2), to which one passes
	 * a file descriptor, the corresponding file is used for 
	 * nothing but looking up the vnode behind (see fsync() 
	 * in kern/vfs_syscalls.c).
	 *
	 * We have thus two choices here: either grab the first
	 * open file and use that for the fsync message or run
	 * through all files (at least, the ones we have acces to,
	 * ie., those belonging to the current process). The first
	 * is bad: as we mentioned concerning sshfs, dirty data can belong
	 * to each ssh messaging thread, being associated to file
	 * handles independently. 
	 *
	 * So we wanna sync all our open files. "Nowait" sync is easy,
	 * just walk over all the files, and send FUSE_FSYNC(DIR) with
	 * a null callback, and leave. "Wait" sync is harder: syncing
	 * the files linearly, one after the other would be very easy
	 * indeed but would be far more than a lethal dose of the KISS
	 * principle. So sending of the messages and waiting for them
	 * should be done synchronously. The messaging backend is flexible
	 * enough for making the implementation of this not too hard, 
	 * but now I don't wanna go mucking with the proper handling of
	 * interrupts, and other issues; and I don't wanna break the nice
	 * hi-level messaging API which serves us well otherwise. So
	 * currently I just say no here...
	 *
	 * Just to note: even the "nowait" type fsync could be optimized
	 * if we went beyond the api, and lock/unlock all mutexes (tickets',
	 * messages', callbacks') only once, and take out/put in stuff for all
	 * files together, with one big breath. But doing not so is not as much
	 * overhead as it would be worth for the trouble.
	 */

	fvdat = vp->v_data;
	/*
	 * It seemed to be a good idea to lock the filehandle lock only before
	 * doing iterate_filehandle but that configuration is suspected to cause
	 * LOR alerts, so now we get the filehandle lock before the fuse_data
	 * lock.
	 */
	sx_slock(&fvdat->fh_lock);
 	if ((err = fdisp_get_vopdata(&fdi, vp->v_mount)))
		goto out;

	iterate_filehandles(vp, td, NULL, 0, fuse_fsync_filehandle, &fdi);  

	sx_sunlock(fdi.slock);

out:
	sx_sunlock(&fvdat->fh_lock);
	return (err);
}


static int
fuse_write_directbackend(struct vnode *vp, uint64_t fh_id, struct uio *uio,
                         struct ucred *cred, struct thread *td)
{	
	struct fuse_dispatcher fdi;
	struct fuse_write_in *fwi;
	size_t chunksize;
	int diff;
	int err = 0;

	if ((err = fdisp_get_vopdata(&fdi, vp->v_mount)))
		return (err);

	while (uio->uio_resid > 0) {
		/* surprise surprise, limit like in linux */
		chunksize = MIN(uio->uio_resid, FUSE_MAX_IN / 2);

		fdi.iosize = sizeof(*fwi) + chunksize;
		fdisp_prepare_msg(&fdi, FUSE_WRITE, fh_id, td, cred);
	
		fwi = fdi.indata;
		fwi->fh = fh_id;
		fwi->offset = uio->uio_offset;
		fwi->size = chunksize;

		if ((err = uiomove((char *)fdi.indata + sizeof(*fwi),
		                   chunksize, uio)))
			break;
	
		if ((err = fdisp_wait_answ(&fdi)))
			return(err);

		diff = chunksize - ((struct fuse_write_out *)fdi.answ)->size;
		if (diff < 0) {
			err = EIO;
			break;
		}

		uio->uio_resid += diff;
		uio->uio_offset -= diff; 
	
		ticket_refresh(fdi.tick);
	}

	ticket_drop(fdi.tick);
	sx_sunlock(fdi.slock);

	return (err);
}

/*
 * Vnode op for write using bio
 */
static int
fuse_write_biobackend(struct vnode *vp, struct uio *uio, struct ucred *cred,
                      struct thread *td)
{
	int biosize;
	struct buf *bp;
	daddr_t lbn;
	int bcount;
	int n, on, err = 0;
	int64_t fsize = vp->v_object->un_pager.vnp.vnp_size;

	DEBUG2G("fsize %lld\n", (long long int)fsize); 

	biosize = vp->v_mount->mnt_stat.f_iosize;
	/*
	 * Find all of this file's B_NEEDCOMMIT buffers.  If our writes
	 * would exceed the local maximum per-file write commit size when
	 * combined with those, we must decide whether to flush,
	 * go synchronous, or return err.  We don't bother checking
	 * IO_UNIT -- we just make all writes atomic anyway, as there's
	 * no point optimizing for something that really won't ever happen.
	 */
	do {
		lbn = uio->uio_offset / biosize;
		on = uio->uio_offset & (biosize-1);
		n = min((unsigned)(biosize - on), uio->uio_resid);
		
		DEBUG2G("lbn %d, on %d, n %d, uio offset %d\n", (int)lbn, on,
		        n, (int)uio->uio_offset);

again:
		/*
		 * Handle direct append and file extension cases, calculate
		 * unaligned buffer size.
		 */

		if (uio->uio_offset == fsize && n) {
			/*
			 * Get the buffer (in its pre-append state to maintain
			 * B_CACHE if it was previously set).  Resize the
			 * nfsnode after we have locked the buffer to prevent
			 * readers from reading garbage.
			 */
			bcount = on;
			DEBUG("getting block from OS, bcount %d\n", bcount);
			bp = getblk(vp, lbn, bcount, PCATCH, 0, 0);

			if (bp != NULL) {
				long save;

				fsize = uio->uio_offset + n;
				vnode_pager_setsize(vp, fsize);

				save = bp->b_flags & B_CACHE;
				bcount += n;
				allocbuf(bp, bcount);
				bp->b_flags |= save;
			}
		} else {
			/*
			 * Obtain the locked cache block first, and then
			 * adjust the file's size as appropriate.
			 */
			bcount = on + n;
			if ((off_t)lbn * biosize + bcount < fsize) {
				if ((off_t)(lbn + 1) * biosize < fsize)
					bcount = biosize;
				else
					bcount = fsize - (off_t)lbn * biosize;
			}
			DEBUG("getting block from OS, bcount %d\n", bcount);
			bp = getblk(vp, lbn, bcount, PCATCH, 0, 0);
			if (uio->uio_offset + n > fsize) {
				fsize = uio->uio_offset + n;
				vnode_pager_setsize(vp, fsize);
			}
		}

		if (!bp) {
			err = EINTR;
			break;
		}

		/*
		 * Issue a READ if B_CACHE is not set.  In special-append
		 * mode, B_CACHE is based on the buffer prior to the write
		 * op and is typically set, avoiding the read.  If a read
		 * is required in special append mode, the server will
		 * probably send us a short-read since we extended the file
		 * on our end, resulting in b_resid == 0 and, thusly,
		 * B_CACHE getting set.
		 *
		 * We can also avoid issuing the read if the write covers
		 * the entire buffer.  We have to make sure the buffer state
		 * is reasonable in this case since we will not be initiating
		 * I/O.  See the comments in kern/vfs_bio.c's getblk() for
		 * more information.
		 *
		 * B_CACHE may also be set due to the buffer being cached
		 * normally.
		 */

		if (on == 0 && n == bcount) {
			bp->b_flags |= B_CACHE;
			bp->b_flags &= ~B_INVAL;
			bp->b_ioflags &= ~BIO_ERROR;
		}

		if ((bp->b_flags & B_CACHE) == 0) {
			bp->b_iocmd = BIO_READ;
			vfs_busy_pages(bp, 0);
			VOP_STRATEGY(vp, bp);
			if ((err =  bp->b_error)) {
				brelse(bp);
				break;
			}
		}
		if (bp->b_wcred == NOCRED)
			bp->b_wcred = crhold(cred);

		/*
		 * If dirtyend exceeds file size, chop it down.  This should
		 * not normally occur but there is an append race where it
		 * might occur XXX, so we log it.
		 *
		 * If the chopping creates a reverse-indexed or degenerate
		 * situation with dirtyoff/end, we 0 both of them.
		 */

		if (bp->b_dirtyend > bcount) {
			DEBUG2G("Fuse append race @%lx:%d\n",
			    (long)bp->b_blkno * biosize,
			    bp->b_dirtyend - bcount);
			bp->b_dirtyend = bcount;
		}

		if (bp->b_dirtyoff >= bp->b_dirtyend)
			bp->b_dirtyoff = bp->b_dirtyend = 0;

		/*
		 * If the new write will leave a contiguous dirty
		 * area, just update the b_dirtyoff and b_dirtyend,
		 * otherwise force a write rpc of the old dirty area.
		 *
		 * While it is possible to merge discontiguous writes due to
		 * our having a B_CACHE buffer ( and thus valid read data
		 * for the hole), we don't because it could lead to
		 * significant cache coherency problems with multiple clients,
		 * especially if locking is implemented later on.
		 *
		 * as an optimization we could theoretically maintain
		 * a linked list of discontinuous areas, but we would still
		 * have to commit them separately so there isn't much
		 * advantage to it except perhaps a bit of asynchronization.
		 */

		if (bp->b_dirtyend > 0 &&
		    (on > bp->b_dirtyend || (on + n) < bp->b_dirtyoff)) {
			/*
 			 * Yes, we mean it. Write out everything to "storage"
 			 * immediatly, without hesitation. (Apart from other
 			 * reasons: the only way to know if a write is valid
 			 * if its actually written out.)
 			 */
			bwrite(bp);
			if (bp->b_error == EINTR) {
				err = EINTR;
				break;
			}
			goto again;
		}

		err = uiomove((char *)bp->b_data + on, n, uio);

		/*
		 * Since this block is being modified, it must be written
		 * again and not just committed.  Since write clustering does
		 * not work for the stage 1 data write, only the stage 2
		 * commit rpc, we have to clear B_CLUSTEROK as well.
		 */
		bp->b_flags &= ~(B_NEEDCOMMIT | B_CLUSTEROK);

		if (err) {
			bp->b_ioflags |= BIO_ERROR;
			bp->b_error = err;
			brelse(bp);
			break;
		}

		/*
		 * Only update dirtyoff/dirtyend if not a degenerate
		 * condition.
		 */
		if (n) {
			if (bp->b_dirtyend > 0) {
				bp->b_dirtyoff = min(on, bp->b_dirtyoff);
				bp->b_dirtyend = max((on + n), bp->b_dirtyend);
			} else {
				bp->b_dirtyoff = on;
				bp->b_dirtyend = on + n;
			}
			vfs_bio_set_validclean(bp, on, n);
		}

		bwrite(bp);
		if ((err =  bp->b_error))
			break;
	} while (uio->uio_resid > 0 && n > 0);

	return (err);
}

static int
fuse_write_f(struct file *fp, struct uio *uio, struct ucred *cred, int flags,
	     struct thread *td)
{
	struct fuse_filehandle *fufh = fp->f_data;
	struct vattr va;
	int err = 0;

	BREAK_IF_BAD(fp);

	if (! FILE_IS_FAT(fp))
		return (EBADF);

	if (uio->uio_resid == 0)
		return (0);

	vn_lock(fp->f_vnode, LK_EXCLUSIVE | LK_RETRY, td);

	if (fp->f_vnode->v_type == VBAD) {
#if _DEBUG2G
		DEBUG2G("bad file\n");
		vn_printf(fp->f_vnode, " * ");
#endif
		VOP_UNLOCK(fp->f_vnode, 0, td);
		return (EBADF);
	}

	if (fp->f_flag & O_APPEND) {
		VOP_GETATTR(fp->f_vnode, &va, cred, td);
		uio->uio_offset = va.va_size;
	} else if ((flags & FOF_OFFSET) == 0)
		uio->uio_offset = fp->f_offset;

	if (fp->f_flag & O_DIRECT ||
	    fufh->flags & FUSEFH_DIRECTIO) {
		DEBUG2G("direct write of vnode %llu via file handle %llu\n",
		        VTOI32(fp->f_vnode), (unsigned long long)fufh->fh_id);
		err = fuse_write_directbackend(fp->f_vnode, fufh->fh_id, uio,
		                               cred, td);
	} else {
		DEBUG2G("buffered write of vnode %llu\n", VTOI32(fp->f_vnode));
		err = fuse_write_biobackend(fp->f_vnode, uio, cred, td);
	}

	if ((flags & FOF_OFFSET) == 0)
		fp->f_offset = uio->uio_offset;
	fp->f_nextoff = uio->uio_offset;
	VOP_UNLOCK(fp->f_vnode, 0, td);

	DEBUG("leaving with %d\n", err);
	return (err);
}

/*
 * Do an I/O operation to/from a cache block. This may be called
 * synchronously or from an nfsiod.
 */
static int
fuse_strategy(struct vop_strategy_args *ap)
{
	return fuse_strategy_i(ap->a_vp, ap->a_bp, NULL);
}

static int
fuse_strategy_i(struct vnode *vp, struct buf *bp, struct fuse_filehandle *fufh)
{
	struct fuse_dispatcher fdi;
	struct ucred *cred; 
	int err = 0;
	int chunksize, respsize;
	caddr_t bufdat;
	int biosize = vp->v_mount->mnt_stat.f_iosize;

#if _DEBUG
	DEBUG("backtrace:\n");
	kdb_backtrace();
#endif	

	if (! (vp->v_type == VREG || vp->v_type == VDIR)) {
		DEBUG("for vnode #%llu v_type is %d, dropping\n",
		      VTOI32(vp), vp->v_type);
		return (EOPNOTSUPP);
	}

	if (bp->b_iocmd != BIO_READ && bp->b_iocmd != BIO_WRITE) {
		DEBUG("for vnode #%llu bio tried with biocmd 0x%x, dropping\n",
		       VTOI32(vp), bp->b_iocmd);
		return (EOPNOTSUPP);
	}

	/*
	 * clear BIO_ERROR and B_INVAL state prior to initiating the I/O. We
	 * do this here so we do not have to do it in all the code that
	 * calls us.
	 */
	bp->b_flags &= ~B_INVAL;
	bp->b_ioflags &= ~BIO_ERROR;

	KASSERT(!(bp->b_flags & B_DONE),
	        ("fuse_strategy: bp %p already marked done", bp));

	if (bp->b_bcount == 0)
		return (0);

	cred = bp->b_iocmd == BIO_READ ? bp->b_rcred : bp->b_wcred;

#if _DEBUG
	DEBUG2G("reading from block #%d at vnode:\n", (int)bp->b_blkno);
	vn_printf(vp, " * ");
#endif
	if (fufh) {
		DEBUG2G("we have a useable filehandle passed on\n");
		fufh->useco++;
	} else if (! (fufh = get_filehandle(vp, NULL, cred,
		              bp->b_iocmd == BIO_READ ? FREAD : FWRITE))) {
		DEBUG2G("fetching filehandle failed\n");
		err = EIO;
	}

	if (! err)
	 	err = fdisp_get_vopdata(&fdi, vp->v_mount);

	if (err)
		goto out;

	DEBUG2G("vp #%llu, fufh #%llu\n", VTOI32(vp), (unsigned long long)fufh->fh_id);

	if (bp->b_iocmd == BIO_READ) {
		struct fuse_read_in *fri;

		bufdat = bp->b_data;
		bp->b_resid = bp->b_bcount;
		while (bp->b_resid > 0) {
			DEBUG2G("starting bio with resid %ld\n", bp->b_resid);	
			chunksize = MIN(bp->b_resid,
			                PAGE_SIZE * FUSE_MAX_PAGES_PER_REQ);
			fdi.iosize = sizeof(*fri);
			fdisp_prepare_msg(&fdi, vp->v_type == VREG ? FUSE_READ : FUSE_READDIR,
			                  VTOI(vp), curthread, cred);
		
			fri = fdi.indata;
			fri->fh = fufh->fh_id;
			fri->offset = ((off_t)bp->b_blkno) * biosize;
			fri->size = chunksize;
			fdi.tick->callbn.resp_bufdata = bufdat;
		
			DEBUG("waiting for answer\n");
			if ((err = fdisp_wait_answ(&fdi)))
				goto out;

			respsize = fdi.tick->callbn.resp_bufsize;
			DEBUG2G("chunksize %d, respsize %d, bp->b_resid %ld, bp->b_bcount %ld\n",
			         chunksize, respsize, bp->b_resid, bp->b_bcount);
			bp->b_resid -= respsize;
			bufdat += respsize;

			DEBUG2G("%d bytes asked for from offset %llu, passing on the %d we got\n",
			        chunksize, (long long unsigned)fri->offset, respsize);
	
			if (respsize < chunksize) {
#if ZERO_PAD_INCOMPLETE_BUFS
				/*
				 * "if we don't get enough data, just fill the
				 * rest with zeros."
				 * in nfs context this means a hole in the
				 * file, I don't know what to do with this
				 * here... [we just get a buch of zeroes
				 * instead of EOF, baaad]
				 */ 
				DEBUG("zeroing tail of %ld bytes\n",
				      bp->b_resid);
				bzero((char *)bp->b_data + bp->b_bcount - bp->b_resid,
				      bp->b_resid);
				bp->b_resid = 0;
#else
				break;
#endif
			}
			ticket_refresh(fdi.tick);
			DEBUG("bp->b_data %p\n", bp->b_data);
		}
	} else {
		struct fuse_write_in *fwi;
		struct fuse_write_out *fwo;
		int diff;
		int merr = 0;

		bufdat = bp->b_data + bp->b_dirtyoff;
		while (bp->b_dirtyend > bp->b_dirtyoff) {
			/* surprise surprise, limit like in linux */
			chunksize = MIN(bp->b_dirtyend - bp->b_dirtyoff,
			                FUSE_MAX_IN / 2);
	
			fdi.iosize = sizeof(*fwi);
			fdisp_prepare_msg(&fdi, FUSE_WRITE, VTOI(vp),
			                  curthread, cred);
		
			fwi = fdi.indata;
			fwi->fh = fufh->fh_id;
			fwi->offset = (off_t)bp->b_blkno * biosize + bp->b_dirtyoff;
			fwi->size = chunksize;
			fdi.tick->msgn.msg_bufdata = bufdat;
			fdi.tick->msgn.msg_bufsize = chunksize;
	
			if ((err = fdisp_wait_answ(&fdi))) {
				merr = 1;
				break;
			}
	
			fwo = fdi.answ;
			diff = chunksize - fwo->size;
			if (diff < 0) {
				err = EIO;
				break;
			}
	
			bp->b_dirtyoff += fwo->size; 
		
			ticket_refresh(fdi.tick);
		}

		if (bp->b_dirtyend == bp->b_dirtyoff)
			bp->b_dirtyend = bp->b_dirtyoff = 0;

		bp->b_resid = bp->b_dirtyend - bp->b_dirtyoff;

		if (merr)
			goto out;
	}

	ticket_drop(fdi.tick);
	sx_sunlock(fdi.slock);

out:
	if (fufh)
		fufh->useco--;

	if (err) {
		bp->b_ioflags |= BIO_ERROR;
		bp->b_error = err;
	}
	DEBUG("calling bufdone\n");      
	bufdone(bp);

	/* 
	 * This is a dangerous function. If returns error, that might mean a
	 * panic. We prefer pretty much anything over being forced to panic by
	 * a malicious daemon (a demon?). So we just return 0 anyway. You
	 * should never mind this: this function has its own error propagation
	 * mechanism via the argument buffer, so not-that-melodramatic
	 * residents of the call chain still will be able to know what to do.
	 */
	return (0);
}	

int
fuse_bmap(struct vop_bmap_args *ap)
{
	DEBUG("a_bn: %d\n", (int)ap->a_bn);

	if (ap->a_bop != NULL)
		*ap->a_bop = &ap->a_vp->v_bufobj;
	if (ap->a_bnp != NULL)
		/* Conversion is identical */
		*ap->a_bnp = ap->a_bn;
	if (ap->a_runp != NULL)
		*ap->a_runp = 0;
	if (ap->a_runb != NULL)
		*ap->a_runb = 0;
	return (0);
}

int
fuse_print(struct vop_print_args *ap)
{
	struct fuse_vnode_data *fvdat = ap->a_vp->v_data;

	printf("nodeid: %llu, fh_counter: %d, nlookup: %llu\n",
	       VTOI32(ap->a_vp), fvdat->fh_counter, (unsigned long long)fvdat->nlookup);
	return (0);
}

static void
fuse_bufstrategy(struct bufobj *bo, struct buf *bp)
{
	struct vnode *vp;
	int rc;

	DEBUG("vnode #%llu\n", VTOI32((struct vnode *)bo->bo_private));
	vp = bo->bo_private;
	KASSERT(bo == &vp->v_bufobj, ("BO/VP mismatch: vp %p (#%llu) bo %p != %p",
	    vp, VTOI32(vp), &vp->v_bufobj, bo));
	rc = VOP_STRATEGY(vp, bp);
	KASSERT(vp->v_op == &dead_vnodeops || rc == 0,
	        ("Fuse VOP_STRATEGY failed: bp=%p, "
	        "vp=%p, rc=%d", bp, vp, rc));
}








/******************************
 *
 * >>> Module management stuff
 *
 ******************************/

/*   Modeled after tunclone() of net/if_tun.c
 */
#if USE_OLD_CLONEHANDLER_API
static void
fusedev_clone(void *arg, char *name, int namelen, struct cdev **dev)
#else
static void
fusedev_clone(void *arg, struct ucred *cred, char *name, int namelen,
              struct cdev **dev)
#endif
{
	/*
	 * Why cloning? We do need per-open info, but we could as well put our
	 * hands on the file struct assigned to an open by implementing
	 * d_fdopen instead of d_open.
	 * 
	 * From that on, the usual way to per-open (that is, file aware)
	 * I/O would be pushing our preferred set of ops into the f_op
	 * field of that file at open time. But that wouldn't work in
	 * FreeBSD, as the devfs open routine (which is the one who calls
	 * the device's d_(fd)open) overwrites that f_op with its own
	 * file ops mercilessly.
	 *
	 * Yet... even if we could get devfs to keep our file ops intact,
	 * I'd still say cloning is better. It makes fuse daemons' identity
	 * explicit and globally visible to userspace, and we are not forced
	 * to get the mount done by the daemon itself like in linux (where 
	 * I/O is file aware by default). (The possibilities of getting the
	 * daemon do the mount or getting the mount util spawn the daemon
	 * are still open, of course; I guess I will go for the latter 
	 * appcocroach.)
	 */

	int i, unit;

	if (*dev != NULL)
		return;

	if (strcmp(name, "fuse") == 0) {
		unit = -1;
	} else if (dev_stdclone(name, NULL, "fuse", &unit) != 1) {
		return;
	}

	/* find any existing device, or allocate new unit number */
	i = clone_create(&fuseclones, &fuse_cdevsw, &unit, dev, 0);
	if (i) {
		*dev = make_dev(&fuse_cdevsw, unit2minor(unit),
			        UID_ROOT, GID_OPERATOR,
		                S_IRUSR|S_IWUSR|S_IRGRP|S_IWGRP,
                                "fuse%d", unit);
		if (*dev != NULL) {
			dev_ref(*dev);
			(*dev)->si_drv1 = NULL;	
			(*dev)->si_flags |= SI_CHEAPCLONE;
		}
	}

	DEBUG("clone done: %d\n",unit);
}

#if FMASTER
extern struct cdevsw fmaster_cdevsw[5];
static struct cdev *fmaster_dev[5];
#endif
extern struct vfsconf fuse_vfsconf;


static void
fuse_bringdown(eventhandler_tag eh_tag)
{
#if FMASTER
	int i;
#endif

	EVENTHANDLER_DEREGISTER(dev_clone, eh_tag);

	clone_cleanup(&fuseclones);
#if USE_FUSE_LOCK
	mtx_destroy(&fuse_mtx);
#endif
#if FMASTER
	for (i = 0; i < 5; i++) {
		DEBUG("destroying fmaster%d\n", i);	
		destroy_dev(fmaster_dev[i]);
	}
#endif
}

static int
fuse_loader(struct module *m, int what, void *arg)
{
	static eventhandler_tag  eh_tag = NULL;
	int err = 0;
#if FMASTER
	int i;
	char *fmaster_name = "fmasterx";
#endif

	GIANT_REQUIRED;

	switch (what) {
	case MOD_LOAD:                /* kldload */

		fuse_fileops.fo_read     = fuse_read_f;
		fuse_fileops.fo_write    = fuse_write_f; 
		fuse_fileops.fo_ioctl    = vnops.fo_ioctl;
		fuse_fileops.fo_poll     = vnops.fo_poll;
		fuse_fileops.fo_kqfilter = vnops.fo_kqfilter;
		fuse_fileops.fo_stat     = vnops.fo_stat;
		fuse_fileops.fo_close    = fuse_close_f;
		fuse_fileops.fo_flags    = DFLAG_PASSABLE | DFLAG_SEEKABLE;

		clone_setup(&fuseclones);
#if USE_FUSE_LOCK
		mtx_init(&fuse_mtx, "fuse_mtx", NULL, MTX_DEF);
#endif
		eh_tag = EVENTHANDLER_REGISTER(dev_clone, fusedev_clone, 0,
		                               1000);
		if (eh_tag == NULL) {
			clone_cleanup(&fuseclones);
#if USE_FUSE_LOCK
			mtx_destroy(&fuse_mtx);
#endif
			return (ENOMEM);
		}
#if FMASTER
		for (i=0; i<5; i++) {
			fmaster_name[7] = i + '0';
			fmaster_dev[i] = make_dev(&fmaster_cdevsw[i], 0,
                                                  UID_ROOT, GID_WHEEL,
						  S_IWUSR, fmaster_name);
		}
#endif
		/* Duh, it's static... */
		/* vfs_register(&fuse_vfsconf); */

		/* vfs_modevent ignores its first arg */
		if ((err = vfs_modevent(NULL, what, &fuse_vfsconf)))
			fuse_bringdown(eh_tag);

		break;
	case MOD_UNLOAD:
		KASSERT(fuse_useco >= 0,
		        ("fuse_useco is negative: %d", fuse_useco));
		if (fuse_useco > 0) {
			DEBUG2G("fuse_useco %d\n", fuse_useco);
		        return (EBUSY);
		}

		if ((err = vfs_modevent(NULL, what, &fuse_vfsconf)))
			return (err);

		/*
		 * at this point the counter falls below zero thus new init
		 * attempts will know that no brownie for them
		 */
		fuse_useco--;

		fuse_bringdown(eh_tag);

		/* vfs_unregister(&fuse_vfsconf); */

		break;
	default:
		return (EINVAL);
	}

	printf("fuse module %sloaded\n",
	       what == MOD_LOAD ? "" : "un");

	return (err);
}

/* Registering the module */

static struct vfsconf fuse_vfsconf = {
	.vfc_version = VFS_VERSION,
	.vfc_name = "fusefs",
	.vfc_vfsops = &fuse_vfsops,
	.vfc_typenum = -1,
	.vfc_flags = VFCF_SYNTHETIC
};

static moduledata_t fuse_moddata = {
	"fuse",
	fuse_loader,
	&fuse_vfsconf
};               	
DECLARE_MODULE(fuse, fuse_moddata, SI_SUB_VFS, SI_ORDER_MIDDLE);

/*
 * >>> The end
 */
