#include <sys/types.h>
#include <sys/param.h>
#include <sys/systm.h>
#include <sys/sysctl.h>
#include <sys/proc.h>
#include <sys/ucred.h>
#include <sys/mac.h>
#include <sys/jail.h>

#define FUSE_AUX 1
#include "fuse.h"

int
get_unprivileged_proc_debug(struct thread *td)
{
	int err;
	int unprivileged_proc_debug;

	err = kernel_sysctlbyname(td, "security.bsd.unprivileged_proc_debug",
		NULL, 0, &unprivileged_proc_debug, sizeof(unprivileged_proc_debug), NULL, 0);

	if (err)
		unprivileged_proc_debug = 0;

	return (unprivileged_proc_debug);
}

/*
 * A modification of p_candebug() of kern/kern_prot.c which does its checks again 
 * a cred (and not a proc)
 */
int
cr_candebug(int unprivileged_proc_debug, struct ucred *dcr, struct ucred *ucr)
{
	int error, grpsubset, i, uidsubset;

	/*
	 * We don't wanna kill the time here if it's as simple as
	 * this 
	 */
	if (dcr->cr_ruid == ucr->cr_ruid)
		return (0);

	if (!unprivileged_proc_debug) {
		error = suser_cred(dcr, SUSER_ALLOWJAIL);
		if (error)
			return (error);
	}
	if ((error = prison_check(dcr, ucr)))
		return (error);
/*
	if ((error = cr_seeotheruids(dcr, ucr)))
		return (error);
	if ((error = cr_seeothergids(dcr, ucr)))
		return (error);
*/
	/*
	 * Is ucr's group set a subset of dcr's effective group set?  This
	 * includes ucr's egid, group access list, rgid, and svgid.
	 */
	grpsubset = 1;
	for (i = 0; i < ucr->cr_ngroups; i++) {
		if (!groupmember(ucr->cr_groups[i], dcr)) {
			grpsubset = 0;
			break;
		}
	}
	grpsubset = grpsubset &&
	    groupmember(ucr->cr_rgid, dcr) &&
	    groupmember(ucr->cr_svgid, dcr);

	/*
	 * Are the uids present in ucr's credential equal to dcr's
	 * effective uid?  This includes ucr's euid, svuid, and ruid.
	 */
	uidsubset = (dcr->cr_uid == ucr->cr_uid &&
	    dcr->cr_uid == ucr->cr_svuid &&
	    dcr->cr_uid == ucr->cr_ruid);

#if 0
	/*
	 * Has the credential of the process changed since the last exec()?
	 */
	credentialchanged = (p->p_flag & P_SUGID);
#endif

	/*
	 * If ucr's gids aren't a subset, or the uids aren't a subset,
	 * or the credential has changed, require appropriate ucrrivilege
	 * for dcr to debug ucr.  For POSIX.1e capabilities, this will
	 * require CAP_SYS_PTRACE.
	 */
	if (!grpsubset || !uidsubset /* || credentialchanged */) {
		error = suser_cred(dcr, SUSER_ALLOWJAIL);
		if (error)
			return (error);
	}

	return (0);
}

#if _DEBUG || _DEBUG2G || FMASTER
static char *pptable[] = { "\\000", "\\001", "\\002", "\\003", "\\004", "\\005", "\\006", "\\a", "\\010", "\\t", "\\n", "\\v", "\\f", "\\r", "\\016", "\\017", "\\020", "\\021", "\\022", "\\023", "\\024", "\\025", "\\026", "\\027", "\\030", "\\031", "\\032", "\\e", "\\034", "\\035", "\\036", "\\037", " ", "!", "\"", "#", "$", "%", "&", "'", "(", ")", "*", "+", ",", "-", ".", "/", "0", "1", "2", "3", "4", "5", "6", "7", "8", "9", ":", ";", "<", "=", ">", "?", "@", "A", "B", "C", "D", "E", "F", "G", "H", "I", "J", "K", "L", "M", "N", "O", "P", "Q", "R", "S", "T", "U", "V", "W", "X", "Y", "Z", "[", "\\", "]", "^", "_", "`", "a", "b", "c", "d", "e", "f", "g", "h", "i", "j", "k", "l", "m", "n", "o", "p", "q", "r", "s", "t", "u", "v", "w", "x", "y", "z", "{", "|", "}", "~", "\\177", "\\200", "\\201", "\\202", "\\203", "\\204", "\\205", "\\206", "\\207", "\\210", "\\211", "\\212", "\\213", "\\214", "\\215", "\\216", "\\217", "\\220", "\\221", "\\222", "\\223", "\\224", "\\225", "\\226", "\\227", "\\230", "\\231", "\\232", "\\233", "\\234", "\\235", "\\236", "\\237", "\\240", "\\241", "\\242", "\\243", "\\244", "\\245", "\\246", "\\247", "\\250", "\\251", "\\252", "\\253", "\\254", "\\255", "\\256", "\\257", "\\260", "\\261", "\\262", "\\263", "\\264", "\\265", "\\266", "\\267", "\\270", "\\271", "\\272", "\\273", "\\274", "\\275", "\\276", "\\277", "\\300", "\\301", "\\302", "\\303", "\\304", "\\305", "\\306", "\\307", "\\310", "\\311", "\\312", "\\313", "\\314", "\\315", "\\316", "\\317", "\\320", "\\321", "\\322", "\\323", "\\324", "\\325", "\\326", "\\327", "\\330", "\\331", "\\332", "\\333", "\\334", "\\335", "\\336", "\\337", "\\340", "\\341", "\\342", "\\343", "\\344", "\\345", "\\346", "\\347", "\\350", "\\351", "\\352", "\\353", "\\354", "\\355", "\\356", "\\357", "\\360", "\\361", "\\362", "\\363", "\\364", "\\365", "\\366", "\\367", "\\370", "\\371", "\\372", "\\373", "\\374", "\\375", "\\376", "\\377" };

void
uprettyprint(char *buf, size_t len)
{
	int i;

	uprintf("\"");	
	for (i=0; i < len; i++) {
		uprintf("%s",pptable[((uint8_t *)buf)[i]]);
	} 
	uprintf("\"");
}

void
prettyprint(char *buf, size_t len)
{
	int i;

	printf("\"");	
	for (i=0; i < len; i++) {
		printf("%s",pptable[((uint8_t *)buf)[i]]);
	} 
	printf("\"");
}

void
fprettyprint(struct fuse_iov *fiov, size_t dlen)
{
	int i;

	uprintf("\"");	
	for (i=0; i < MIN(fiov->len, dlen); i++) {
		uprintf("%s",pptable[((uint8_t *)fiov->base)[i]]);
	} 
	uprintf("\"");
	uprintf("%s\n", dlen < fiov->len ? "..." : "");
}
#endif
